"""
Module containing a Units library

    :copyright: 2020 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

"""

from collections import defaultdict
from decimal import Decimal, InvalidOperation
from numbers import Number

from math import ceil, floor
from pint import UnitRegistry
from pint.errors import UndefinedUnitError
from pint.quantity import _Quantity
from pint.util import UnitsContainer


def to_decimal(number):
    """
    Casts a number to a Decimal safely.

    Parameters
    ----------
    number: Number
        number to be cast to a decimal

    Returns
    -------
    Decimal
        decimal representation of the input number

    Raises
    ------
    ValueError
        Number is not of a type that's castable to decimal
    """
    if isinstance(number, Decimal):
        decimal = number
    elif isinstance(number, Number):
        decimal = Decimal(str(number))
    else:
        raise ValueError(
            f"Tried to cast {number} to decimal but it was of non-numeric type "
            f"{type(number)}."
        )
    return decimal


# pragma pylint: disable=protected-access
class DecimalUnitRegistry(UnitRegistry):
    """
    Redefines builtin UnitRegistry methods for doing conversions to use Decimals
    instead of floats to eliminate floating point imprecision, particularly
    converting .to("new_unit").
    """

    def _get_root_units(self, input_units, check_nonmult=True):
        if not input_units:
            return Decimal("1"), UnitsContainer()

        # The cache is only done for check_nonmult=True
        if check_nonmult and input_units in self._root_units_cache:
            return self._root_units_cache[input_units]

        accumulators = [Decimal("1"), defaultdict(Decimal)]
        self._get_root_units_recurse(input_units, Decimal("1"), accumulators)

        factor = accumulators[0]
        units = UnitsContainer(
            dict((k, v) for k, v in accumulators[1].items() if v != Decimal("0"))
        )

        if check_nonmult:
            for unit in units.keys():
                if not self._units[unit].converter.is_multiplicative:
                    return None, units

        if check_nonmult:
            self._root_units_cache[input_units] = factor, units

        return factor, units

    def _get_root_units_recurse(self, ref, exp, accumulators):
        for key in sorted(ref):
            exp2 = to_decimal(exp) * to_decimal(ref[key])
            key = self.get_name(key)
            reg = self._units[key]
            if reg.is_base:
                accumulators[1][key] += exp2
            else:
                accumulators[0] *= to_decimal(reg._converter.scale) ** exp2
                if reg.reference is not None:
                    self._get_root_units_recurse(reg.reference, exp2, accumulators)


# Preload UnitRegistry (Use default Pints definition file as a base)
_UnitRegistry = DecimalUnitRegistry()

"""Map string representation of Pint units over to Autoprotocol format"""
# Map Temperature Unit names
_UnitRegistry._units["degC"]._name = "celsius"
_UnitRegistry._units["celsius"]._name = "celsius"
_UnitRegistry._units["degF"]._name = "fahrenheit"
_UnitRegistry._units["fahrenheit"]._name = "fahrenheit"
_UnitRegistry._units["degR"]._name = "rankine"
_UnitRegistry._units["rankine"]._name = "rankine"
# Map Speed Unit names
_UnitRegistry._units["revolutions_per_minute"]._name = "rpm"

"""Add support for Molarity Unit"""
_UnitRegistry.define("molar = mole/liter = M")
# pragma pylint: enable=protected-access


class UnitError(Exception):
    """
    Exceptions from creating new Unit instances with bad inputs.
    """

    message_text = "Unit error for %s"

    def __init__(self, value):
        super(UnitError, self).__init__(self.message_text % value)
        self.value = value


class UnitStringError(UnitError):
    message_text = (
        "Invalid format '%s'; when building a Unit from a string "
        "it must be formatted as '1:meter'."
    )


class UnitValueError(UnitError):
    message_text = "Invalid value '%s'; when building a Unit the value must be numeric."


class UnitUnitsError(UnitError):
    message_text = (
        "Invalid value '%s'; when building a Unit "
        "the units must be in the UnitRegistry."
    )


class Unit(_Quantity):
    """
    A representation of a measure of physical quantities such as length,
    mass, time and volume.
    Uses Pint's Quantity as a base class for implementing units and
    inherits functionalities such as conversions and proper unit
    arithmetic.
    Note that the magnitude is stored as a double-precision float, so
    there are inherent issues when dealing with extremely large/small
    numbers as well as numerical rounding for non-base 2 numbers.

    Example
    -------

        .. code-block:: python

            vol_1 = Unit(10, 'microliter')
            vol_2 = Unit(10, 'liter')
            print(vol_1 + vol_2)

            time_1 = Unit(1, 'second')
            speed_1 = vol_1/time_1
            print (speed_1)
            print (speed_1.to('liter/hour'))

    Returns
    -------
    Unit
        unit object

        .. code-block:: none

            10000010.0:microliter
            10.0:microliter / second
            0.036:liter / hour

    """

    def __new__(cls, value, units=None):
        cls._REGISTRY = _UnitRegistry
        cls.force_ndarray = False

        # Automatically return Unit if Unit is provided
        if isinstance(value, Unit):
            return value

        # Automatically parse String if no units provided
        if not units and isinstance(value, str):
            try:
                value, units = value.split(":")
            except ValueError:
                raise UnitStringError(value)
        try:
            return super(Unit, cls).__new__(cls, Decimal(str(value)), units)
        except (ValueError, InvalidOperation):
            raise UnitValueError(value)
        except UndefinedUnitError:
            raise UnitUnitsError(units)

    def __init__(self, value, units=None):  # pylint: disable=unused-argument
        super(Unit, self).__init__()
        self.unit = self.units.__str__()

    def __str__(self, ndigits=12):
        """
        Parameters
        ----------
        ndigits: int, optional
            Number of decimal places to round to, useful for numerical
            precision reasons

        Returns
        -------
        str
            This rounds the string presentation to 12 decimal places by default
            to account for the majority of numerical precision issues
        """
        rounded_magnitude = round(self.magnitude, ndigits)
        normalized_magnitude = to_decimal(rounded_magnitude).normalize()
        unit_repr = self.unit.replace("**", "^").replace(" ", "")
        return f"{normalized_magnitude:f}:{unit_repr:s}"

    def __repr__(self):
        return f"Unit({self.magnitude:f}, '{self.units:s}')"

    def __ceil__(self):
        return self.__class__(ceil(self.magnitude), self.units)

    def __floor__(self):
        return self.__class__(floor(self.magnitude), self.units)

    def _mul_div(self, other, magnitude_op, units_op=None):
        """
        Extends Pint's base _Quantity multiplication/division
        implementation by checking for dimensionality and
        casting Numbers to Decimals
        """
        if isinstance(other, Unit):
            if self.dimensionality == other.dimensionality:
                other = other.to(self.units)
        else:
            other = to_decimal(other)

        return super(Unit, self)._mul_div(other, magnitude_op, units_op)

    def _imul_div(self, other, magnitude_op, units_op=None):
        """
        Extends Pint's base _Quantity multiplication/division
        implementation by checking for dimensionality and
        casting Numbers to Decimals
        """
        if isinstance(other, Unit):
            if self.dimensionality == other.dimensionality:
                other = other.to(self.units)
        else:
            other = to_decimal(other)

        return super(Unit, self)._imul_div(other, magnitude_op, units_op)

    @property
    def magnitude(self):
        return self._magnitude

    @magnitude.setter
    def magnitude(self, magnitude):
        try:
            self._magnitude = to_decimal(magnitude)
        except ValueError:
            raise RuntimeError(
                f"Tried to set Unit's magnitude {magnitude} but it was of type "
                f"{type(magnitude)}. Magnitudes must be numeric."
            )

    @staticmethod
    def fromstring(s):
        return Unit(s)

    def ceil(self):
        """
        Equivalent of math.ceil(Unit) for python 2 compatibility

        Returns
        -------
        Unit
            ceil of Unit
        """
        return self.__ceil__()

    def floor(self):
        """
        Equivalent of math.floor(Unit) for python 2 compatibility

        Returns
        -------
        Unit
            floor of Unit
        """
        return self.__floor__()

    def round(self, ndigits):
        """
        Equivalent of round(Unit) for python 2 compatibility

        Parameters
        ----------
        ndigits: int
            number of decimal places to be rounded to

        Returns
        -------
        Unit
            rounded unit
        """
        return self.__round__(ndigits)



"""Mix LiquidHandleMethod

Base LiquidHandleMethod used by Protocol.mix to generate a series of
movements within individual wells.

    :copyright: 2020 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details
"""
from .transfer import LiquidHandleMethod
from ..instruction import LiquidHandle
from ..unit import Unit
from ..util import parse_unit


# pylint: disable=protected-access
class Mix(LiquidHandleMethod):
    """LiquidHandleMethod for generating transfers within wells

    LiquidHandleMethod for moving volume within a single well.

    Attributes
    ----------
    _liquid : LiquidClass
        used to determine calibration, flowrates, and sensing thresholds

    Notes
    -----
    The primary entry points that for this class are:
        - _mix_transports : generates transports within a single location

    See Also
    --------
    LiquidHandleMethod : base LiquidHandleMethod with reused functionality
    Protocol.mix : the standard interface for interacting with Mix
    """

    def __init__(self, tip_type=None, blowout=True, repetitions=None, position_z=None):
        """
        Parameters
        ----------
        tip_type : str, optional
            tip_type to be used for the LiquidHandlingMethod
        blowout : bool or dict, optional
            whether to execute a blowout step or the parameters for one.
            this generates two operations, an initial air aspiration before
            entering any wells, and a corresponding final air dispense
            after the last operation that involves liquid
            See Also LiquidHandle.builders.blowout
        repetitions : int
            the number of times the mix should be repeated
        position_z : dict
            the position that the tip should move to prior to mixing, if the
            position references the `liquid_surface` then mix movements will
            track the surface with the defined offset.
            See Also LiquidHandle.builders.position_z
        """
        super(Mix, self).__init__(tip_type=tip_type, blowout=blowout)

        # parameters for required behavior
        self.repetitions = repetitions
        self.position_z = position_z

        # LiquidHandle parameters that are generated and modified at runtime
        self._liquid = None

    def _has_calibration(self):
        liquids = [self._liquid]
        return any(_ and _._has_calibration() for _ in liquids)

    def _calculate_overage_volume(self, volume):
        calibration_overage = (
            self._estimate_calibrated_volume(volume, self._liquid, self.tip_type)
            - volume
        )

        return calibration_overage

    def default_blowout(self, volume):
        return LiquidHandle.builders.blowout(
            volume=Unit("5:ul"),
            initial_z=self.default_well_top_position_z(),
            flowrate=None,
        )

    def _mix_transports(self, volume):
        """Generates mix transports

        Generates and returns all of the transports that should happen within
        the well of a mix operation.

        Calls a series of _transport_`y` helper methods that each query the `y`
        parameter and default_`y` method to decide on a set of behavior and use
        that to define transports that are appended to the _transports list.

        Parameters
        ----------
        volume : Unit

        Notes
        -----
        This method defines what lower level transport-generating methods are
        called and in what order. It can be overwritten when adding an
        entirely new set of transport-generating behavior.

        Return
        ------
        list
            transports corresponding to the mix operation
        """
        self._transports = []
        volume = parse_unit(volume, "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_pre_buffer(volume)
        self._transport_mix(volume)
        self._transport_blowout(volume)

        return self._transports

    def _transport_mix(self, volume):
        """Mixes the volume

        Parameters
        ----------
        volume : Unit

        Raises
        ------
        TypeError
            if mix repetitions is not an int

        See Also
        --------
        repetitions : holds any user defined repetition parameters
        default_repetitions : specifies default repetition parameters
        position_z : holds any user defined position_z parameters
        default_position_z : specifies default position_z parameters
        _mix : lower level helper that generates the mix transports
        """
        repetitions = self.repetitions or self.default_repetitions(volume)
        position_z = self.position_z or self.default_position_z(volume)

        if not isinstance(repetitions, int):
            raise TypeError(f"Mix repetitions {repetitions} was not an int.")
        position_z = LiquidHandle.builders.position_z(**position_z)
        self._mix(
            volume=volume,
            repetitions=repetitions,
            initial_z=position_z,
            asp_flowrate=self._liquid._get_aspirate_flowrate(volume, self.tip_type),
            dsp_flowrate=self._liquid._get_dispense_flowrate(volume, self.tip_type),
            delay_time=self._liquid.delay_time,
            liquid_class=self._liquid.name,
        )

    # pylint: disable=unused-argument
    def default_repetitions(self, volume):
        """Default mix repetitions

        Parameters
        ----------
        volume: Unit

        Returns
        -------
        int
            number of mix repetitions

        See Also
        --------
        repetitions : holds any user defined repetition parameters
        _transport_mix : generates the actual mix transports
        """
        return 10

    # pylint: disable=unused-argument
    def default_position_z(self, volume):
        """Default position_z

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            mix position_z

        See Also
        --------
        position_z : holds any user defined position_z parameters
        _transport_mix : generates the actual mix transports
        """
        if self._is_single_channel():
            position_z = self.default_lld_position_z(self._liquid)
        else:
            position_z = self.default_well_bottom_position_z()
        return position_z



"""
Liquid handling utilities and methods utilized by Protocol.transfer to
generate LiquidHandle instructions.
"""
from .liquid_class import LiquidClass
from .liquid_handle_method import LiquidHandleMethod
from .mix import *
from .transfer import *



"""Builders
Module containing builders, which help build inputs for Instruction parameters

    :copyright: 2020 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

Summary
-------
These builder methods are used to generate and validate complex data
structures used in Autoprotocol specification. Each of them is capable of
using their own output as input. Therefore these builders are also used as
inline checks in Protocol methods.

Notes
-----
Generally these builders should not be called from this file directly.
They're more easily accessible by referencing a specific Instruction's
builders attribute (e.g. `Spectrophotometry.Builders.mode_params`).

See Also
--------
Instruction
    Instructions corresponding to each of the builders
"""

from collections import defaultdict
from numbers import Number

from collections.abc import Iterable  # pylint: disable=no-name-in-module
from functools import reduce

from .constants import SBS_FORMAT_SHAPES
from .container import WellGroup, Well, Container
from .unit import Unit
from .util import parse_unit, is_valid_well


class InstructionBuilders(object):  # pylint: disable=too-few-public-methods
    """General builders that apply to multiple instructions
    """

    def __init__(self):
        self.sbs_shapes = ["SBS96", "SBS384"]

    @staticmethod
    def _merge_param_dicts(left=None, right=None):
        """Finds the union of two dicts of params and checks for duplicates

        Parameters
        ----------
        left : dict or None
            Parameters to be merged
        right : dict or None
            Parameters to be merged

        Returns
        -------
        dict
            A merged set of parameters from the left and right dicts

        Raises
        ------
        ValueError
            if multiple values are specified for the same parameter

        """
        left = left or dict()
        right = right or dict()

        union = defaultdict(list)
        for params in (left, right):
            for key, value in params.items():
                if value is not None:
                    union[key].append(value)

        unique = dict()
        for key, value in union.items():
            if len(value) == 1:
                unique[key] = value[0]
            else:
                raise ValueError(
                    f"Parameter: {key} had multiple values: {value} specified."
                )

        return unique

    # pylint: disable=redefined-builtin
    def shape(self, rows=1, columns=1, format=None):
        """
        Helper function for building a shape dictionary

        Parameters
        ----------
        rows : int, optional
            Number of rows to be concurrently transferred
        columns : int, optional
            Number of columns to be concurrently transferred
        format : str, optional
            Plate format in String form. e.g. "SBS96" or "SBS384"

        Returns
        -------
        dict
            shape parameters

        Raises
        ------
        TypeError
            If rows/columns aren't ints
        ValueError
            If an invalid row/column count is given
        ValueError
            If an invalid shape is given
        ValueError
            If rows/columns are greater than what is allowed for the format
        """
        if not isinstance(rows, int) or not isinstance(columns, int):
            raise TypeError("Rows/columns have to be of type integer")

        if format is None:
            for shape in self.sbs_shapes:
                valid_rows = rows <= SBS_FORMAT_SHAPES[shape]["rows"]
                valid_columns = columns <= SBS_FORMAT_SHAPES[shape]["columns"]
                if valid_rows and valid_columns:
                    format = shape
                    break

            if not format:
                raise ValueError("Invalid number of rows and/or columns specified")

        if format not in self.sbs_shapes:
            raise ValueError(
                f"Invalid shape format; format has to be in {self.sbs_shapes}"
            )

        valid_rows = rows <= SBS_FORMAT_SHAPES[format]["rows"]
        valid_columns = columns <= SBS_FORMAT_SHAPES[format]["columns"]
        if not (valid_rows and valid_columns):
            raise ValueError(
                f"rows: {rows} and columns: {columns} are not possible with "
                f"format: {format}."
            )

        return {"rows": rows, "columns": columns, "format": format}


class ThermocycleBuilders(InstructionBuilders):
    """
    These builders are meant for helping to construct the `groups`
    argument in the `Protocol.thermocycle` method
    """

    def __init__(self):
        super(ThermocycleBuilders, self).__init__()
        self.valid_dyes = {
            "FAM",
            "SYBR",  # channel 1
            "VIC",
            "HEX",
            "TET",
            "CALGOLD540",  # channel 2
            "ROX",
            "TXR",
            "CALRED610",  # channel 3
            "CY5",
            "QUASAR670",  # channel 4
            "QUASAR705",  # channel 5
            "FRET",  # channel 6
        }

    def dyes(self, **kwargs):
        """Helper function for creating a dye parameter

        Parameters
        ----------
        **kwargs : dict(str: int or list(int))
            A mapping from a dye (str) to the index of a well

        Returns
        -------
        dict
            A thermocycling dye to well mapping

        Raises
        ------
        ValueError
            If any of the specified dyes are not valid
        ValueError
            If wells is not an int, str, list(int), or list(str)
        """
        dyes = {}
        for dye, wells in kwargs.items():
            if dye not in self.valid_dyes:
                raise ValueError(
                    f"dye {dye} is not in the set of valid dyes " f"{self.valid_dyes}"
                )
            if not isinstance(wells, list):
                wells = [wells]
            if not all(isinstance(_, (int, str)) for _ in wells):
                raise ValueError(
                    f"dye {dye} had wells {wells} that were not an int, str " "or list"
                )
            dyes[dye] = wells

        return dyes

    def dyes_from_well_map(self, well_map):
        """Helper function for creating a dye parameter from a well_map

        Take a map of wells to the dyes it contains and returns a map of dyes to
        the list of wells that contain it.

        Parameters
        ----------
        well_map : dict(well, str)
            A thermocycling well to dye mapping

        Returns
        -------
        dict
            A thermocycling dye to well mapping

        See Also
        --------
        Thermocycle.Builders.dyes : standard constructor for the dyes parameter

        """
        dyes = reduce(lambda x, y: x.union(y), [set(well_map[k]) for k in well_map])

        dye_map = {dye: [] for dye in dyes}
        for well in well_map:
            dyes = well_map[well]
            for dye in dyes:
                dye_map[dye] += [well]

        return self.dyes(**dyes)

    @staticmethod
    def melting(start=None, end=None, increment=None, rate=None):
        """Helper function for creating melting parameters

        Generates melt curve parameters for Thermocycle Instructions.

        Parameters
        ----------
        start : str or Unit
            The starting temperature for the melt curve
        end : str or Unit
            The ending temperature for the melt curve
        increment : str or Unit
            The temperature increment of the melt curve
        rate : str or Unit
            The duration the individual increments

        Returns
        -------
        dict
            A thermocycling melt curve specification

        Raises
        ------
        ValueError
            If some, but not all melt curve parameters are specified

        """
        melting_params = [start, end, increment, rate]

        if any(melting_params) and not all(melting_params):
            raise ValueError(
                "To specify a melt curve, you must specify values for "
                "start, end, increment, and rate"
            )

        if start:
            start = parse_unit(start, "celsius")
        if end:
            end = parse_unit(end, "celsius")
        if increment:
            increment = parse_unit(increment, "celsius")
        if rate:
            rate = parse_unit(rate, "second")

        return {"start": start, "end": end, "increment": increment, "rate": rate}

    def group(self, steps, cycles=1):
        """
        Helper function for creating a thermocycle group, which is a series of
        steps repeated for the number of cycles

        Parameters
        ----------
        steps: list(ThermocycleBuilders.step)
            Steps to be carried out. At least one step has to be specified.
            See `ThermocycleBuilders.step` for more information
        cycles: int, optional
            Number of cycles to repeat the specified steps. Defaults to 1

        Returns
        -------
        dict
            A thermocycling group

        Raises
        ------
        TypeError
            Invalid input types, i.e. `cycles` is not of type int and `steps`
            is not of type list
        ValueError
            `cycles` is not positive
        ValueError
            `steps` does not contain any elements
        """
        if not isinstance(cycles, int):
            raise TypeError(f"`cycles` {cycles} has to be of type int")
        if not isinstance(steps, list):
            raise TypeError(f"`steps` {steps} has to be of type list")

        if cycles <= 0:
            raise ValueError(f"`cycles` {cycles} has to be positive")
        if not steps:
            raise ValueError("`steps` has to contain at least one element")

        # Reformatting to use temperature for gradient input
        def reformat_gradient(**kwargs):
            if "gradient" in kwargs:
                kwargs["temperature"] = kwargs.pop("gradient")
            return kwargs

        group_dict = dict(
            cycles=cycles, steps=[self.step(**reformat_gradient(**_)) for _ in steps]
        )

        return group_dict

    @staticmethod
    def step(temperature, duration, read=None):
        """
        Helper function for creating a thermocycle step.

        Parameters
        ----------
        temperature: Unit or dict(str, Unit)
            Block temperature which the contents should be thermocycled at.

            If a gradient thermocycle is desired, specifying a dict with
            "top" and "bottom" keys will control the desired temperature
            at the top and bottom rows of the block, creating a gradient
            along the column.

            ..code-block:: python

              temperature = {"top": "50:celsius", "bottom": "45:celsius"}

        duration: str or Unit
            Duration where the specified temperature parameters will be applied
        read: Boolean, optional
            Determines if a read at wavelengths specified by the dyes in the
            parent `thermocycle` instruction will be enabled for this particular
            step. Useful for qPCR applications.

        Returns
        -------
        dict
            A thermocycling step

        Raises
        ------
        TypeError
            Invalid input types, e.g. `read` is not of type bool
        ValueError
            Invalid format specified for `temperature` dict
        ValueError
            Duration is not greater than 0 second

        """
        step_dict = dict()
        if isinstance(temperature, dict):
            if set(temperature.keys()) != {"top", "bottom"}:
                raise ValueError(
                    f"{temperature} was specified, but only 'top' and 'bottom' "
                    f"keys are allowed for a temperature dictionary"
                )
            step_dict["gradient"] = dict(
                top=parse_unit(temperature["top"], "celsius"),
                bottom=parse_unit(temperature["bottom"], "celsius"),
            )
        else:
            step_dict["temperature"] = parse_unit(temperature, "celsius")

        duration = parse_unit(duration, "second")
        if duration <= Unit("0:second"):
            raise ValueError("Step `duration` has to be at least 1 second")
        step_dict["duration"] = duration

        if read is not None:
            if not isinstance(read, bool):
                raise TypeError(f"`read` {read} has to be of type bool")
            step_dict["read"] = read

        return step_dict


class SPEBuilders(InstructionBuilders):
    """
    These builders are meant for helping to construct arguments for the
    `SPE` instruction.
    """

    def spe_params(self, params, is_elute=False):
        if not isinstance(params, list):
            raise ValueError("SPE mobile phase parameters {} must be a list.")
        parsed_params = []
        for param in params:
            parsed_params.append(self.mobile_phase_params(is_elute=is_elute, **param))
        return parsed_params

    def mobile_phase_params(
        self,
        volume,
        loading_flowrate,
        settle_time,
        processing_time,
        flow_pressure,
        resource_id=None,
        is_sample=False,
        destination_well=None,
        is_elute=False,
    ):
        """
        Create a dictionary with mobile phase parameters which can be
        used as input for instructions.

        Parameters
        ----------
        volume: str or Unit
            the duration to shake the plate for
        loading_flowrate: str or Unit, optional
            amplitude of shaking between 1 and 6:millimeter
        settle_time: bool, optional
            True for orbital and False for linear shaking
        processing_time: str or Unit
            Duration for which pressure is applied to the cartridge
            after `settle_time` has elapsed.
        flow_pressure: str or Unit
            Pressure applied to the column.
        resource_id: str
            Resource ID of desired solvent.
        is_sample: bool
            If a sample is processed.
        destination_well: Well
            Destination well for eluate.  Required parameter for
            each `elute` mobile phase parameter
        is_elute: bool
            If an elute is processed.

        Returns
        -------
        dict
            spe mobile_phase_params

        Raises
        ------
        ValueError
            if shake `flow_pressure` is not positive
        ValueError
            if mobile phase solvent and resource_id not included
        TypeError
            if `resource_id` is not a string
        ValueError
            if `settle_time` is not positive
        ValueError
            if `processing_time` is not positive

        """
        volume = parse_unit(volume, "milliliter")
        settle_time = parse_unit(settle_time, "second")
        if settle_time <= Unit(0, "second"):
            raise ValueError(f"settle_time: {settle_time} is not positive")

        processing_time = parse_unit(processing_time, "second")
        if processing_time <= Unit(0, "second"):
            raise ValueError(f"processing_time: {processing_time} is not " "positive")

        flow_pressure = parse_unit(flow_pressure, "bar")
        if flow_pressure <= Unit(0, "bar"):
            raise ValueError(f"flow_pressure: {flow_pressure} must be " f"positive.")
        loading_flowrate = parse_unit(loading_flowrate, "ul/s")

        if not is_sample:
            if not resource_id:
                raise ValueError(
                    "A 'resource_id' must be included " "for mobile_phase_params."
                )
            if not isinstance(resource_id, str):
                raise TypeError("'resource_id' {} must be a string")
        if is_elute:
            if not destination_well:
                raise ValueError(
                    "A 'destination_well' must be included " "for elute parameter."
                )
            if not isinstance(destination_well, Well):
                raise TypeError("'destination_well' {} must be of type Well")

        mode_params = {
            "volume": volume,
            "settle_time": settle_time,
            "processing_time": processing_time,
            "flow_pressure": flow_pressure,
            "loading_flowrate": loading_flowrate,
            "resource_id": resource_id,
            "destination_well": destination_well,
        }

        mode_params = {k: v for k, v in mode_params.items() if v is not None}

        return mode_params


class DispenseBuilders(InstructionBuilders):
    """
    These builders are meant for helping to construct arguments in the
    `Protocol.dispense` method.
    """

    def __init__(self):
        super(DispenseBuilders, self).__init__()
        self.SHAKE_PATHS = ["landscape_linear"]

    @staticmethod
    # pragma pylint: disable=unused-argument
    def nozzle_position(position_x=None, position_y=None, position_z=None):
        """
        Generates a validated nozzle_position parameter.

        Parameters
        ----------
        position_x : Unit, optional
        position_y : Unit, optional
        position_z : Unit, optional

        Returns
        -------
        dict
            Dictionary of nozzle position parameters
        """

        position_dict = {
            name: parse_unit(position, "mm")
            for name, position in locals().items()
            if position is not None
        }

        return position_dict

    # pragma pylint: enable=unused-argument

    @staticmethod
    def column(column, volume):
        """
        Generates a validated column parameter.

        Parameters
        ----------
        column : int
        volume : str, Unit

        Returns
        -------
        dict
            Column parameter of type {"column": int, "volume": Unit}

        """
        return {"column": int(column), "volume": parse_unit(volume, "uL")}

    def columns(self, columns):
        """
        Generates a validated columns parameter.

        Parameters
        ----------
        columns : list({"column": int, "volume": str, Unit})

        Returns
        -------
        list
            List of columns of type ({"column": int, "volume": str, Unit})

        Raises
        ------
        ValueError
            No `column` specified for columns
        ValueError
            Non-unique column indices
        """
        if not len(columns) > 0:
            raise ValueError("There must be at least one column specified for columns.")

        column_list = [self.column(**_) for _ in columns]

        if len(column_list) != len(set([_["column"] for _ in column_list])):
            raise ValueError(
                f"Column indices must be unique, but there were duplicates "
                f"in {column_list}."
            )

        return column_list

    def shake_after(self, duration, frequency=None, path=None, amplitude=None):
        """
        Generates a validated shake_after parameter.

        Parameters
        ----------
        duration : Unit, str
        frequency : Unit, str, optional
        path : str, optional
        amplitude : Unit, str, optional

        Returns
        -------
        dict
            Shake after dictionary of type {"duration": Unit,
            "frequency": Unit, "path": str, "amplitude": Unit}

        Raises
        ------
        ValueError
            Invalid shake path specified
        """

        if path and path not in self.SHAKE_PATHS:
            raise ValueError(
                f"Invalid shake path {path} specified, must be one of "
                f"{self.SHAKE_PATHS}"
            )

        shake_after = {
            "duration": parse_unit(duration, "seconds"),
            "frequency": parse_unit(frequency, "hertz") if frequency else None,
            "path": path,
            "amplitude": parse_unit(amplitude, "mm") if amplitude else None,
        }

        return {k: v for k, v in shake_after.items() if v is not None}


class SpectrophotometryBuilders(InstructionBuilders):
    """
    These builders are meant for helping to construct arguments for the
    `Spectrophotometry` instruction.
    """

    def __init__(self):
        super(SpectrophotometryBuilders, self).__init__()
        self.MODES = {
            "absorbance": self.absorbance_mode_params,
            "fluorescence": self.fluorescence_mode_params,
            "luminescence": self.luminescence_mode_params,
            "shake": self.shake_mode_params,
        }

        self.READ_POSITIONS = ["top", "bottom"]

        self.SHAKE_PATHS = [
            "portrait_linear",
            "landscape_linear",
            "cw_orbital",
            "ccw_orbital",
            "portrait_down_double_orbital",
            "landscape_down_double_orbital",
            "portrait_up_double_orbital",
            "landscape_up_double_orbital",
            "cw_diamond",
            "ccw_diamond",
        ]

        self.Z_REFERENCES = ["plate_bottom", "plate_top", "well_bottom", "well_top"]

        self.Z_HEURISTICS = [
            "max_mean_read_without_saturation",
            "closest_distance_without_saturation",
        ]

    @staticmethod
    def wavelength_selection(shortpass=None, longpass=None, ideal=None):
        """
        Generates a representation of a wavelength selection by either
        filters (using shortpass/longpass) or monochromators (using ideal)

        Parameters
        ----------
        shortpass : Unit, str, optional
        longpass : Unit, str, optional
        ideal : Unit, str, optional

        Returns
        -------
        dict
            Wavelength selection parameters.
        """

        selection = {
            "shortpass": parse_unit(shortpass, "nanometer") if shortpass else None,
            "longpass": parse_unit(longpass, "nanometer") if longpass else None,
            "ideal": parse_unit(ideal, "nanometer") if ideal else None,
        }

        selection = {k: v for k, v in selection.items() if v is not None}

        return selection

    def groups(self, groups):
        """
        Parameters
        ----------
        groups : list(dict)
            A list of spectrophotometry groups.

        Returns
        -------
        list(dict)
            A list of spectrophotometry groups.
        """
        return [self.group(_["mode"], _["mode_params"]) for _ in groups]

    def group(self, mode, mode_params):
        """
        Parameters
        ----------
        mode : str
            A string representation of a valid spectrophotometry mode.
        mode_params : dict
            A dict of mode_params corresponding to the mode.

        Returns
        -------
        dict
            A spectrophotometry group.

        Raises
        ------
        ValueError
            Invalid mode specified
        """
        if mode not in self.MODES.keys():
            raise ValueError(
                f"Invalid mode {mode}, must be in valid modes " f"{self.MODES.keys()}."
            )

        return {"mode": mode, "mode_params": self.MODES[mode](**mode_params)}

    def absorbance_mode_params(
        self,
        wells,
        wavelength,
        num_flashes=None,
        settle_time=None,
        read_position=None,
        position_z=None,
    ):
        """
        Parameters
        ----------
        wells : iterable(Well) or WellGroup
            Wells to be read.
        wavelength : Unit or str
            The wavelengths at which to make absorbance measurements.
        num_flashes : int, optional
            The number of discrete reads to be taken and then averaged.
        settle_time : Unit or str, optional
            The time to wait between moving to a well and reading it.
        read_position: Enum("top", "bottom"), optional
            The position of the probe relative to the plate for the read
        position_z: dict, optional
            This should be specified with either `position_z_manual` or
            `position_z_calculated`

        Returns
        -------
        dict
            Formatted mode_params for an absorbance mode.

        Raises
        ------
        TypeError
            Invalid type specified for input parameters, e.g. `num_flashes`
            not of type int
        ValueError
            Invalid wells specified
        """
        if not is_valid_well(wells):
            raise ValueError(
                f"Invalid wells {wells}, must be an iterable of wells or a "
                f"WellGroup."
            )

        if isinstance(wells, Well):
            wells = WellGroup([wells])

        if not isinstance(wavelength, list):
            wavelength = [wavelength]

        wavelength = [parse_unit(_, "nanometer") for _ in wavelength]

        if num_flashes is not None and not isinstance(num_flashes, int):
            raise TypeError(f"Invalid num_flashes {num_flashes}, must be an int")

        if settle_time is not None:
            settle_time = parse_unit(settle_time, "second")

        if read_position is not None and read_position not in self.READ_POSITIONS:
            raise ValueError(
                f"Invalid read_position {read_position}, must be in "
                f"{self.READ_POSITIONS}."
            )

        if position_z is not None:
            position_z = self._position_z(position_z)

        mode_params = {
            "wells": wells,
            "wavelength": wavelength,
            "num_flashes": num_flashes,
            "settle_time": settle_time,
            "read_position": read_position,
            "position_z": position_z,
        }

        mode_params = {k: v for k, v in mode_params.items() if v is not None}

        return mode_params

    def fluorescence_mode_params(
        self,
        wells,
        excitation,
        emission,
        num_flashes=None,
        settle_time=None,
        lag_time=None,
        integration_time=None,
        gain=None,
        read_position=None,
        position_z=None,
    ):
        """
        Parameters
        ----------
        wells : iterable(Well) or WellGroup
            Wells to be read.
        excitation : list(dict)
            A list of SpectrophotometryBuilders.wavelength_selection to
            determine the wavelegnth(s) of excitation light used.
        emission : list(dict)
            A list of SpectrophotometryBuilders.wavelength_selection to
            determine the wavelegnth(s) of emission light used.
        num_flashes : int, optional
            The number of discrete reads to be taken and then combined.
        settle_time : Unit or str, optional
            The time to wait between moving to a well and reading it.
        lag_time : Unit or str, optional
            The time to wait between excitation and reading.
        integration_time : Unit or str, optional
            Time over which the data should be collected and integrated.
        gain : int, optional
            The amount of gain to be applied to the readings.
        read_position : str, optional
            The position from which the wells should be read.
        position_z: dict, optional
            This should be specified with either `position_z_manual` or
            `position_z_calculated`

        Returns
        -------
        dict
            Formatted mode_params for a fluorescence mode.

        Raises
        ------
        TypeError
            Invalid input types, e.g. settle_time is not of type Unit(second)
        ValueError
            Invalid wells specified
        ValueError
            Gain is not between 0 and 1
        """
        if not is_valid_well(wells):
            raise ValueError(
                f"Invalid wells {wells}, must be an iterable of wells or a "
                f"WellGroup."
            )

        if isinstance(wells, Well):
            wells = WellGroup([wells])

        if not isinstance(excitation, list):
            raise ValueError("Excitation {} must be a list")
        if not isinstance(emission, list):
            raise ValueError("Emission {} must be a list")

        excitation = [self.wavelength_selection(**_) for _ in excitation]
        emission = [self.wavelength_selection(**_) for _ in emission]

        if num_flashes is not None and not isinstance(num_flashes, int):
            raise ValueError(f"Invalid num_flashes {num_flashes}, must be an int")

        if settle_time is not None:
            settle_time = parse_unit(settle_time, "second")

        if lag_time is not None:
            lag_time = parse_unit(lag_time, "second")

        if integration_time is not None:
            integration_time = parse_unit(integration_time, "second")

        if gain is not None:
            if not isinstance(gain, (int, float)):
                raise TypeError(f"Invalid gain {gain}, must be an int")
            gain = float(gain)
            if not 0 <= gain <= 1:
                raise ValueError(
                    f"Invalid gain {gain}, must be between 0 and 1 (inclusive)."
                )

        if read_position is not None and read_position not in self.READ_POSITIONS:
            raise ValueError(
                f"Invalid read_position {read_position}, must be in "
                f"{self.READ_POSITIONS}."
            )

        if position_z is not None:
            position_z = self._position_z(position_z)

        mode_params = {
            "wells": wells,
            "excitation": excitation,
            "emission": emission,
            "num_flashes": num_flashes,
            "settle_time": settle_time,
            "lag_time": lag_time,
            "integration_time": integration_time,
            "gain": gain,
            "read_position": read_position,
            "position_z": position_z,
        }

        mode_params = {k: v for k, v in mode_params.items() if v is not None}

        return mode_params

    def luminescence_mode_params(
        self,
        wells,
        num_flashes=None,
        settle_time=None,
        integration_time=None,
        gain=None,
        read_position=None,
        position_z=None,
    ):
        """
        Parameters
        ----------
        wells : iterable(Well) or WellGroup
            Wells to be read.
        num_flashes : int, optional
            The number of discrete reads to be taken and then combined.
        settle_time : Unit or str, optional
            The time to wait between moving to a well and reading it.
        integration_time : Unit or str, optional
            Time over which the data should be collected and integrated.
        gain : int, optional
            The amount of gain to be applied to the readings.
        read_position: Enum("top", "bottom"), optional
            The position of the probe relative to the plate for the read
        position_z: dict, optional
            This should be specified with either `position_z_manual` or
            `position_z_calculated`

        Returns
        -------
        dict
            Formatted mode_params for a luminescence mode.

        Raises
        ------
        TypeError
            Invalid input types, e.g. settle_time is not of type Unit(second)
        ValueError
            Gain is not between 0 and 1

        """
        if not is_valid_well(wells):
            raise ValueError(
                f"Invalid wells {wells}, must be an iterable of wells or a "
                f"WellGroup."
            )

        if isinstance(wells, Well):
            wells = WellGroup([wells])

        if num_flashes is not None and not isinstance(num_flashes, int):
            raise TypeError(f"Invalid num_flashes {num_flashes}, must be an int")

        if settle_time is not None:
            settle_time = parse_unit(settle_time, "second")

        if integration_time is not None:
            integration_time = parse_unit(integration_time, "second")

        if gain is not None:
            if not isinstance(gain, (int, float)):
                raise TypeError(f"Invalid gain {gain}, must be an int")
            gain = float(gain)
            if not 0 <= gain <= 1:
                raise ValueError(
                    f"Invalid gain {gain}, must be between 0 and 1 (inclusive)."
                )

        if read_position is not None and read_position not in self.READ_POSITIONS:
            raise ValueError(
                f"Invalid read_position {read_position}, must be in "
                f"{self.READ_POSITIONS}."
            )

        if position_z is not None:
            position_z = self._position_z(position_z)

        mode_params = {
            "wells": wells,
            "num_flashes": num_flashes,
            "settle_time": settle_time,
            "integration_time": integration_time,
            "gain": gain,
            "read_position": read_position,
            "position_z": position_z,
        }

        mode_params = {k: v for k, v in mode_params.items() if v is not None}

        return mode_params

    def shake_mode_params(
        self, duration=None, frequency=None, path=None, amplitude=None
    ):
        """
        Parameters
        ----------
        duration : Unit or str, optional
            The duration of the shaking incubation, if not specified then the
            incubate will last until the end of read interval.
        frequency : Unit or str, optional
            The frequency of the shaking motion.
        path : str, optional
            The name of a shake path. See the spectrophotometry ASC for
            diagrams of different shake paths.
        amplitude : Unit or str, optional
            The amplitude of the shaking motion.

        Returns
        -------
        dict
            Formatted mode_params for a shake mode.
        """
        return self._shake(
            duration=duration, frequency=frequency, path=path, amplitude=amplitude
        )

    def shake_before(self, duration, frequency=None, path=None, amplitude=None):
        """
        Parameters
        ----------
        duration : Unit or str
            The duration of the shaking incubation.
        frequency : Unit or str, optional
            The frequency of the shaking motion.
        path : str, optional
            The name of a shake path. See the spectrophotometry ASC for
            diagrams of different shake paths.
        amplitude : Unit or str, optional
            The amplitude of the shaking motion.

        Returns
        -------
        dict
            Formatted mode_params for a shake mode.
        """
        duration = parse_unit(duration, "second")

        return self._shake(
            duration=duration, frequency=frequency, path=path, amplitude=amplitude
        )

    def _shake(self, duration=None, frequency=None, path=None, amplitude=None):
        """
        Helper method for validating shake params.
        """
        if duration is not None:
            duration = parse_unit(duration, "second")

        if frequency is not None:
            frequency = parse_unit(frequency, "hertz")

        if path and path not in self.SHAKE_PATHS:
            raise ValueError(
                f"Invalid read_position {path}, must be in {self.SHAKE_PATHS}."
            )

        if amplitude is not None:
            amplitude = parse_unit(amplitude, "millimeter")

        params = {
            "duration": duration,
            "frequency": frequency,
            "path": path,
            "amplitude": amplitude,
        }

        params = {k: v for k, v in params.items() if v is not None}

        return params

    def position_z_manual(self, reference=None, displacement=None):
        """Helper for building position_z parameters for a manual position_z
        configuration

        Parameters
        ----------
        reference : str, optional
            Must be one of "plate_top", "plate_bottom", "well_top",
            "well_bottom"
        displacement: Unit or str, optional
            Displacement from reference position.
            Negative would refer to the `well_top` to `well_bottom` direction,
            while positive would refer to the opposite direction.

        Returns
        -------
        dict
            position_z parameters for a Spectrophotometry instruction

        Raises
        ------
        ValueError
            If reference was not in the allowed list
        ValueError
            If invalid displacement was provided
        """
        if reference is not None and reference not in self.Z_REFERENCES:
            raise ValueError(
                f"reference must be one of {self.Z_REFERENCES} but {reference} "
                f"was specified"
            )
        if displacement is not None:
            displacement = parse_unit(displacement, "mm")

        return {"manual": {"reference": reference, "displacement": displacement}}

    def position_z_calculated(self, wells, heuristic=None):
        """Helper for building position_z parameters for a calculated
        position_z configuration

        Parameters
        ----------
        wells : list(Well)
            List of wells to calculate the z-position from
        heuristic: str, optional
            Must be one of "max_mean_read_without_saturation" or
            "closest_distance_without_saturation".
            Please refer to `ASC-041 <http://autoprotocol.org/ascs/#ASC-040>`_
            for the full explanation

        Returns
        -------
        dict
            position_z parameters for a Spectrophotometry instruction

        Raises
        ------
        ValueError
            If a list of wells is not provided
        ValueError
            If an invalid heuristic is specified

        """
        if any([not is_valid_well(well) for well in wells]):
            raise ValueError("Only an iterable of wells is allowed")

        if heuristic is not None and heuristic not in self.Z_HEURISTICS:
            raise ValueError(
                f"heuristic must be one of {self.Z_HEURISTICS} but {heuristic} "
                f"was specified"
            )

        return {"calculated_from_wells": {"wells": wells, "heuristic": heuristic}}

    def _position_z(self, position_z):
        """
        Helper method for validating position_z params
        """
        suggested_msg = (
            "Please use either `position_z_manual` or "
            "`position_z_calculated` functions to construct "
            "the appropriate z-position."
        )

        if not isinstance(position_z, dict):
            raise TypeError(
                f"Invalid position_z {position_z}, must be a dict. " f"{suggested_msg}"
            )

        if "calculated_from_wells" in position_z:
            return self.position_z_calculated(**position_z["calculated_from_wells"])
        elif "manual" in position_z:
            return self.position_z_manual(**position_z["manual"])
        else:
            raise ValueError(
                f"Invalid position_z {position_z} specified. {suggested_msg}"
            )


class LiquidHandleBuilders(InstructionBuilders):
    """Builders for LiquidHandle Instructions
    """

    def __init__(self):
        super(LiquidHandleBuilders, self).__init__()
        self.liquid_classes = ["air", "default", "viscous"]
        self.xy_max = 1
        self.z_references = [
            "well_top",
            "well_bottom",
            "liquid_surface",
            "preceding_position",
        ]
        self.z_detection_methods = ["capacitance", "pressure", "tracked"]
        self.dispense_modes = ["air_displacement", "positive_displacement"]

    def location(self, location=None, transports=None):
        """Helper for building locations

        Parameters
        ----------
        location : Well or str, optional
            Location refers to the well location where the transports will be
            carried out
        transports : list(dict), optional
            Transports refer to the list of transports that will be carried out
            in the specified location
            See Also LiquidHandle.builders.transport

        Returns
        -------
        dict
            location parameters for a LiquidHandle instruction

        Raises
        ------
        TypeError
            If locations aren't str/well
        ValueError
            If transports are specified, but empty
        """
        if not (location is None or isinstance(location, (Well, str))):
            raise TypeError(f"Location {location} is not of type str or Well")

        if transports is not None:
            if not isinstance(transports, Iterable):
                raise ValueError(f"Transports: {transports} is not iterable")
            transports = [self.transport(**_) for _ in transports]
            if len(transports) < 1:
                raise ValueError(
                    f"transports {transports} must be nonempty if specified"
                )

        return {"location": location, "transports": transports}

    def transport(
        self,
        volume=None,
        pump_override_volume=None,
        flowrate=None,
        delay_time=None,
        mode_params=None,
        density=None,
    ):
        """Helper for building transports

        Parameters
        ----------
        volume : Unit or str, optional
            Volume to be aspirated/dispensed. Positive volume -> Dispense.
            Negative -> Aspirate
        pump_override_volume : Unit or str, optional
            Calibrated volume, volume which the pump will move
        flowrate : dict, optional
            Flowrate parameters
            See Also LiquidHandle.builders.flowrate
        delay_time : Unit or str, optional
            Time spent waiting after executing mandrel and pump movement
        mode_params : dict, optional
            Mode parameters
            See Also LiquidHandle.builders.mode_params
        density : Unit or str, optional
            Density of the liquid to be aspirated/dispensed in mg/ml.


        Returns
        -------
        dict
            transport parameters for a LiquidHandle instruction
        """
        if volume is not None:
            volume = parse_unit(volume, "ul")
        if density is not None:
            density = parse_unit(density, "mg/ml")
        if pump_override_volume is not None:
            pump_override_volume = parse_unit(pump_override_volume, "ul")
        if flowrate is not None:
            flowrate = self.flowrate(**flowrate)
        if delay_time is not None:
            delay_time = parse_unit(delay_time, "s")
        if mode_params is not None:
            mode_params = self.mode_params(**mode_params)

        return {
            "volume": volume,
            "density": density,
            "pump_override_volume": pump_override_volume,
            "flowrate": flowrate,
            "delay_time": delay_time,
            "mode_params": mode_params,
        }

    @staticmethod
    def flowrate(
        target, initial=None, cutoff=None, acceleration=None, deceleration=None
    ):
        """Helper for building flowrates

        Parameters
        ----------
        target : Unit or str
            Target flowrate
        initial : Unit or str, optional
            Initial flowrate
        cutoff : Unit or str, optional
            Cutoff flowrate
        acceleration : Unit or str, optional
            Volumetric acceleration for initial to target (in ul/s^2)
        deceleration : Unit or str, optional
            Volumetric deceleration for target to cutoff (in ul/s^2)

        Returns
        -------
        dict
            flowrate parameters for a LiquidHandle instruction
        """
        target = parse_unit(target, "ul/s")
        if initial is not None:
            initial = parse_unit(initial, "ul/s")
        if cutoff is not None:
            cutoff = parse_unit(cutoff, "ul/s")
        if acceleration is not None:
            acceleration = parse_unit(acceleration, "ul/s/s")
        if deceleration is not None:
            deceleration = parse_unit(deceleration, "ul/s/s")

        return {
            "target": target,
            "initial": initial,
            "cutoff": cutoff,
            "acceleration": acceleration,
            "deceleration": deceleration,
        }

    def mode_params(
        self,
        liquid_class=None,
        position_x=None,
        position_y=None,
        position_z=None,
        tip_position=None,
    ):
        """Helper for building transport mode_params

        Mode params contain information about tip positioning and the
        liquid being manipulated

        Parameters
        ----------
        liquid_class : str, optional
            The name of the liquid class to be handled. This affects how
            vendors handle populating liquid handling defaults.
        position_x : dict, optional
            Target relative x-position of tip in well.
            See Also LiquidHandle.builders.position_xy
        position_y : dict, optional
            Target relative y-position of tip in well.
            See Also LiquidHandle.builders.position_xy
        position_z : dict, optional
            Target relative z-position of tip in well.
            See Also LiquidHandle.builders.position_z
        tip_position : dict, optional
            A dict of positions x, y, and z. Should only be specified if none of
            the other tip position parameters have been specified.

        Returns
        -------
        dict
            mode_params for a LiquidHandle instruction

        Raises
        ------
        ValueError
            If liquid_class is not in the allowed list
        ValueError
            If both position_x|y|z and tip_position are specified
        """
        if liquid_class is not None and liquid_class not in self.liquid_classes:
            raise ValueError(
                f"liquid_class must be one of {self.liquid_classes} but "
                f"{liquid_class} was specified"
            )

        tip_position = self._merge_param_dicts(
            dict(position_x=position_x, position_y=position_y, position_z=position_z),
            tip_position,
        )
        position_x = tip_position.get("position_x")
        position_y = tip_position.get("position_y")
        position_z = tip_position.get("position_z")

        if position_x is not None:
            position_x = self.position_xy(**position_x)

        if position_y is not None:
            position_y = self.position_xy(**position_y)

        if position_z is not None:
            position_z = self.position_z(**position_z)

        return {
            "liquid_class": liquid_class,
            "tip_position": {
                "position_x": position_x,
                "position_y": position_y,
                "position_z": position_z,
            },
        }

    @staticmethod
    def move_rate(target=None, acceleration=None):
        """Helper for building move_rates

        Parameters
        ----------
        target : Unit or str, optional
            Target velocity. Must be in units of
        acceleration : Unit or str, optional
            Acceleration. Must be in units of

        Returns
        -------
        dict
            move_rate parameters for a LiquidHandle instruction
        """
        if target is not None:
            target = parse_unit(target, "mm/s")
        if acceleration is not None:
            acceleration = parse_unit(acceleration, "mm/s/s")

        return {"target": target, "acceleration": acceleration}

    def position_xy(self, position=None, move_rate=None):
        """Helper for building position_x and position_y parameters

        Parameters
        ----------
        position : Numeric, optional
            Target relative x/y-position of tip in well in unit square
            coordinates.
        move_rate : dict, optional
            The rate at which the tip moves in the well
            See Also LiquidHandle.builders.move_rate

        Returns
        -------
        dict
            position_xy parameters for a LiquidHandle instruction

        Raises
        ------
        TypeError
            If position is non-numeric
        ValueError
            If position is not in range
        """
        if not (position is None or isinstance(position, (float, int))):
            raise TypeError(f"position {position} is not of type float/int")
        if not (position is None or -self.xy_max <= position <= self.xy_max):
            raise ValueError(
                f"position {position} was not in range {-self.xy_max} - "
                f"{self.xy_max}"
            )

        if move_rate is not None:
            move_rate = self.move_rate(**move_rate)

        return {"position": position, "move_rate": move_rate}

    def position_z(
        self,
        reference=None,
        offset=None,
        move_rate=None,
        detection_method=None,
        detection_threshold=None,
        detection_duration=None,
        detection_fallback=None,
        detection=None,
    ):
        """Helper for building position_z parameters

        Parameters
        ----------
        reference : str, optional
            Must be one of "well_top", "well_bottom", "liquid_surface",
             "preceding_position"
        offset : Unit or str, optional
            Offset from reference position
        move_rate : dict, optional
            Controls the rate at which the tip moves in the well
            See Also LiquidHandle.builders.move_rate
        detection_method : str, optional
            Must be one of "tracked", "pressure", "capacitance"
        detection_threshold : Unit or str, optional
            The threshold which must be crossed before a positive reading is
            registered. This is applicable for capacitance and pressure
            detection methods
        detection_duration : Unit or str, optional
            The contiguous duration where the threshold must be crossed before a
            positive reading is registered.
            This is applicable for pressure detection methods
        detection_fallback : dict, optional
            Fallback option which will be used if sensing fails
            See Also LiquidHandle.builders.position_z
        detection : dict, optional
            A dict of detection parameters. Should only be specified if none of
            the other detection parameters have been specified.

        Returns
        -------
        dict
            position_z parameters for a LiquidHandle instruction

        Raises
        ------
        ValueError
            If reference was not in the allowed list
        ValueError
            If both detection_method|duration|threshold|fallback and
            detection are specified
        ValueError
            If detection_method is not in the allowed list
        ValueError
            If detection parameters were specified, but the reference
            position doesn't support detection
        """
        if reference is not None and reference not in self.z_references:
            raise ValueError(
                f"reference must be one of {self.z_references} but "
                f"{reference} was specified"
            )
        if offset is not None:
            offset = parse_unit(offset, "mm")
        if move_rate is not None:
            move_rate = self.move_rate(**move_rate)

        detection = self._merge_param_dicts(
            dict(
                method=detection_method,
                threshold=detection_threshold,
                duration=detection_duration,
                fallback=detection_fallback,
            ),
            detection,
        )

        if any(detection.values()) and not reference == "liquid_surface":
            raise ValueError(
                f"detection parameters were specified, but reference "
                f"{reference} does not support detection"
            )

        method = detection.get("method")
        duration = detection.get("duration")
        threshold = detection.get("threshold")
        fallback = detection.get("fallback")

        if method is not None and method not in self.z_detection_methods:
            raise ValueError(
                f"detection_method must be one of {self.z_detection_methods} "
                f"but {method} was specified"
            )

        if duration is not None:
            duration = parse_unit(duration, "s")

        if threshold is not None:
            threshold = parse_unit(threshold, ["pascal", "farad"])

        if fallback is not None:
            fallback = self.position_z(**fallback)

        detection["method"] = method
        detection["duration"] = duration
        detection["threshold"] = threshold
        detection["fallback"] = fallback

        return {
            "reference": reference,
            "offset": offset,
            "move_rate": move_rate,
            "detection": detection,
        }

    @staticmethod
    def instruction_mode_params(tip_type=None):
        """Helper for building instruction mode_params

        Parameters
        ----------
        tip_type : str, optional
            the string representation ofa tip_type
            See Also tip_type.py

        Returns
        -------
        dict
            mode_params for a LiquidHandle instruction
        """

        return {"tip_type": tip_type}

    def mix(self, volume, repetitions, initial_z, asp_flowrate=None, dsp_flowrate=None):
        """Helper for building mix params for Transfer LiquidHandleMethods

        Parameters
        ----------
        volume : Unit or str
            the volume of the mix step
        repetitions : int
            the number of times that the mix should be repeated
        initial_z : dict
            the position that the tip should move to prior to mixing
            See Also LiquidHandle.builders.position_z
        asp_flowrate : dict, optional
            the flowrate of the aspiration portions of the mix
            See Also LiquidHandle.builders.flowrate
        dsp_flowrate : dict, optional
            the flowrate of the dispense portions of the mix
            See Also LiquidHandle.builders.flowrate

        Returns
        -------
        dict
            mix parameters for a LiquidHandleMethod

        Raises
        ------
        TypeError
            If repetitions is not an int
        """
        volume = parse_unit(volume, "ul")
        if not isinstance(repetitions, int):
            raise TypeError(f"repetitions {repetitions} is not an int")
        initial_z = self.position_z(**initial_z)
        if asp_flowrate is not None:
            asp_flowrate = self.flowrate(**asp_flowrate)
        if dsp_flowrate is not None:
            dsp_flowrate = self.flowrate(**dsp_flowrate)

        return {
            "volume": volume,
            "repetitions": repetitions,
            "initial_z": initial_z,
            "asp_flowrate": asp_flowrate,
            "dsp_flowrate": dsp_flowrate,
        }

    def blowout(self, volume, initial_z, flowrate=None):
        """Helper for building blowout params for LiquidHandleMethods

        Parameters
        ----------
        volume : Unit or str
            the volume of the blowout step
        initial_z : dict
            the position that the tip should move to prior to blowing out
            See Also LiquidHandle.builders.position_z
        flowrate : dict, optional
            the flowrate of the blowout
            See Also LiquidHandle.builders.flowrate

        Returns
        -------
        dict
            blowout params for a LiquidHandleMethod
        """
        volume = parse_unit(volume, "ul")
        initial_z = self.position_z(**initial_z)
        if flowrate is not None:
            flowrate = self.flowrate(**flowrate)

        return {"volume": volume, "initial_z": initial_z, "flowrate": flowrate}

    def desired_mode(self, transports=None, mode=None):
        """Helper for selecting dispense mode based on liquid_class name
        For non-viscous, water-like liquid and air, the method will default
        to "air_displacement". To allow for more accurate aspirate and
        dispense volumes, 'viscous' liquid where air pressure alone is
        sometimes not sufficient to overcome the surface tension to pull
        or push the liquid sufficiently through transfer tip, it will
        default to "positive_displacement" unless otherwise specified.

        Examples
        --------

        .. code-block:: python

            example_transports = [
                LiquidHandle.builders.transport(
                    volume=Unit(1, "uL"),
                    density=None,
                    pump_override_volume=Unit(2, "uL"),
                    flowrate=LiquidHandle.builders.flowrate(
                        target=Unit(10, "uL/s")
                    ),
                    delay_time=Unit(0.5, "s"),
                    mode_params=LiquidHandle.builders.mode_params(
                        liquid_class="air",
                        position_z=LiquidHandle.builders.position_z(
                            reference="preceding_position"
                        )
                    )
                ),
                LiquidHandle.builders.transport(
                    volume=Unit(1, "uL"),
                    density=None,
                    pump_override_volume=Unit(2, "uL"),
                    flowrate=LiquidHandle.builders.flowrate(
                        target=Unit(10, "uL/s")
                    ),
                    delay_time=Unit(0.5, "s"),
                    mode_params=LiquidHandle.builders.mode_params(
                        liquid_class="viscous",
                        position_z=LiquidHandle.builders.position_z(
                            reference="preceding_position"
                        )
                    )
                )
            ]

        LiquidHandle.builders.desired_mode(example_transports, None)

        This will return "positive_displacement" based on the "viscous" liquid
        class. Note that "air" (which defaults to "air_displacement") does not
        cause a conflict since "air" is often applied in additional transfers
        of air (such as blowout) for any liquid class, and is disregarded.

        Parameters
        ----------
        transports : dict, optional
            Dictionary of the transport parameters
        mode : str, optional
            Mode of dispense type

        Returns
        -------
        str
            Mode of dispense type

        Raises
        ------
        TypeError
            mode is not in str
        ValueError
            mode does not contain valid mode value
        TypeError
            liquid is not of str
        ValueError
            liquid_class is not a valid value
        ValueError
            liquid_class did not resolve to a mode
        ValueError
            multiple liquid_class exists in one LiquidHandle
        """
        # dict for default dispense mode for each liquid_class
        liquid_class_to_dispense_mode = {
            "air": "air_displacement",
            "default": "air_displacement",
            "viscous": "positive_displacement",
        }
        if mode:
            if mode not in self.dispense_modes:
                raise ValueError(
                    f"mode: {mode} must be one of the valid modes: "
                    f"{self.dispense_modes}"
                )
            return mode

        # get liquid_classes from transport input
        mode_params = [transport["mode_params"] for transport in transports]
        if all(mode_param is None for mode_param in mode_params):
            liquid_classes = {None}
        else:
            liquid_classes = set(
                [mode_param["liquid_class"] for mode_param in mode_params]
            )
        # remove automatically added 'air' (e.g. from blowout steps) and None
        # classes.
        other_classes = liquid_classes - {"air", None}
        if not other_classes:
            return liquid_class_to_dispense_mode["air"]

        modes = set()
        for other_class in other_classes:
            # resolve mode for different other_class classes (except for None).
            if not isinstance(other_class, str):
                raise TypeError(f"liquid_class: {other_class} is not of type str")
            if other_class not in self.liquid_classes:
                raise ValueError(
                    f"liquid_class: {other_class} must be one of the "
                    f"valid flasses: {self.liquid_classes} "
                )
            if other_class not in liquid_class_to_dispense_mode.keys():
                raise ValueError(
                    f"liquid_class: {other_class} did not resolve accordingly. "
                    f"If there is a new liquid_class, make sure dictionary:"
                    f"{liquid_class_to_dispense_mode} is updated."
                )
            modes.add(liquid_class_to_dispense_mode[other_class])
        # return error if modes is empty.
        if not modes:
            raise ValueError(
                f"modes: {modes} resulted in an empty set. Make sure valid "
                f"mode is added for each liquid class."
            )
        # return error if there are incompatible liquid_class a set of
        # transports.
        if len(set(modes)) > 1:
            raise ValueError(
                f"There are multiple liquid_class types which could "
                f"potentially have different modes: {set(modes)}. Please "
                f"specify the mode to be used from: {self.dispense_modes}. "
                f"Only one mode is allowed per transfer."
            )
        return list(modes)[0]


class PlateReaderBuilders(InstructionBuilders):
    """Helpers for building parameters for plate reading instructions
    """

    def incubate_params(
        self, duration, shake_amplitude=None, shake_orbital=None, shaking=None
    ):
        """
        Create a dictionary with incubation parameters which can be used as
        input for instructions. Currently supports plate reader instructions and
        could be extended for use with other instructions.

        Parameters
        ----------
        duration: str or Unit
            the duration to shake the plate for
        shake_amplitude: str or Unit, optional
            amplitude of shaking between 1 and 6:millimeter
        shake_orbital: bool, optional
            True for orbital and False for linear shaking
        shaking : dict, optional
            A dict of amplitude and orbital: should only be specified if none of
            the other tip shake parameters have been specified.
            Dictionary of incubate parameters

        Returns
        -------
        dict
            plate reader incubate_params

        Raises
        ------
        ValueError
            if shake `duration` is not positive
        ValueError
            if only one of shake_amplitude or shake_orbital is set
        TypeError
            if `shake_orbital` is not a bool
        ValueError
            if `shake_amplitude` is not positive

        """
        duration = parse_unit(duration, "second")
        if duration <= Unit(0, "second"):
            raise ValueError(f"duration: {duration} is not positive")

        shaking = self._merge_param_dicts(
            dict(amplitude=shake_amplitude, orbital=shake_orbital), shaking
        )
        amplitude = shaking.get("amplitude")
        orbital = shaking.get("orbital")

        if (amplitude is not None) and (orbital is not None):
            amplitude = parse_unit(amplitude, "millimeter")
            if amplitude <= Unit(0, "millimeter"):
                raise ValueError(f"shake_amplitude: {amplitude} is not positive")
            if not isinstance(orbital, bool):
                raise TypeError(f"shake_orbital: {orbital} is not a bool")
        elif (amplitude is not None) ^ (orbital is not None):
            raise ValueError(
                f"shake_amplitude: {amplitude} and shake_orbital: {orbital} "
                f"must both be specified to shake"
            )

        return {
            "duration": duration,
            "shaking": {"amplitude": amplitude, "orbital": orbital},
        }


class EvaporateBuilders(InstructionBuilders):
    """
    Helpers for building Evaporate instructions
    """

    def __init__(self):
        super(EvaporateBuilders, self).__init__()
        self.valid_modes = ["rotate", "centrifuge", "vortex", "blowdown"]
        self.valid_gases = ["nitrogen", "argon", "helium"]
        self.rotary_params = [
            "flask_volume",
            "rotation_speed",
            "vacuum_pressure",
            "condenser_temperature",
        ]
        self.centrifugal_params = [
            "spin_acceleration",
            "vacuum_pressure",
            "condenser_temperature",
        ]
        self.vortex_params = [
            "vortex_speed",
            "vacuum_pressure",
            "condenser_temperature",
        ]
        self.blowdown_params = ["gas", "blow_rate", "vortex_speed"]

    def get_mode_params(self, mode, mode_params):
        """
        Checks on the validity of mode and mode_params, and
        creates a dictionary for mode_params

        Parameters
        ----------
        mode : Str
            Mode of the evaporate method.
        mode_params : Dict
            Method parameters for each mode.

        Returns
        -------
        Dict
            Dictionary of mode_params

        Raises
        ------
        ValueError
            If mode is not specified.
        ValueError
            If specified mode is not a valid mode.
        ValueError
            If mode_params contain key(s) that are not
            applicable for the specified mode.
        TypeError
            If there are multiple speed parameters.
        ValueError
            If container agitation speed is less than 0.
        ValueError
            If vacuum_pressure is less than 0 torr.

        """
        mode_to_param_dict = {
            "rotary": self.rotary_params,
            "centrifugal": self.centrifugal_params,
            "vortex": self.vortex_params,
            "blowdown": self.blowdown_params,
        }
        if mode not in self.valid_modes:
            raise ValueError(
                f"Specified mode: {mode} is not valid. Make sure it is one "
                f"of {self.valid_modes}"
            )

        mode_param_keys = mode_to_param_dict[mode]

        if set(mode_params.keys()) != set(mode_param_keys):
            raise ValueError(
                f"Specified mode_params: {mode_params.keys()} do not contain "
                f"appropriate key(s) applicable to the mode: {mode}, which "
                f"include {mode_param_keys}"
            )
        speed_unit_dict = {
            "rotation_speed": "rpm",
            "vortex_speed": "rpm",
            "spin_acceleration": "g",
        }
        mode_param_output = {}
        speed_param = [(k) for k in mode_params.keys() if k in speed_unit_dict.keys()]
        if len(speed_param) > 1:
            raise TypeError(f"There are multiple speed parameters: {speed_param}.")

        for s in speed_param:
            if Unit(mode_params[s]).magnitude <= 0:
                raise ValueError(f"{s} is less than or equal to 0.")
            else:
                mode_param_output[s] = parse_unit(mode_params[s], speed_unit_dict[s])

        if "vacuum_pressure" in mode_params.keys():
            pressure = mode_params["vacuum_pressure"]
            if pressure is not None:
                pressure = parse_unit(pressure, "torr")
                if pressure <= Unit("0: torr"):
                    raise ValueError("vacuum_pressure cannot be lower than 0 torr.")
                else:
                    mode_param_output["vacuum_pressure"] = pressure

        if "condenser_temperature" in mode_params.keys():
            temp = mode_params["condenser_temperature"]
            if temp is not None:
                temp = parse_unit(temp, "celsius")
                mode_param_output["condenser_temperature"] = temp

        if "gas" in mode_params.keys():
            if not mode_params["gas"] in self.valid_gases:
                raise TypeError(
                    f"Specified gas: {mode_params['gas']} is not included in "
                    f"valid gases: {self.valid_gases}"
                )
            mode_param_output["gas"] = mode_params["gas"]

        if "blow_rate" in mode_params.keys():
            blow_rate = mode_params["blow_rate"]
            if not blow_rate is None:
                blow_rate = parse_unit(blow_rate, "mL/min")
                mode_param_output["blow_rate"] = blow_rate

        return mode_param_output


class GelPurifyBuilders(InstructionBuilders):
    """Helpers for building GelPurify instructions
    """

    def extract(self, source, band_list, lane=None, gel=None):
        """Helper for building extract params for gel_purify

        Parameters
        ----------
        source : Well
            The Well that contains the sample be purified
        band_list : list(dict)
            A list of bands to be extracted from the source sample
        lane : int, optional
            The gel lane for the source sample to be run on
        gel : int, optional
            The number of the gel if using multiple gels

        Returns
        -------
        dict
            gel_purify extract parameters

        Raises
        ------
        TypeError
            If source is not a Well

        """
        if not isinstance(source, Well):
            raise TypeError(f"source: {source} is not a Well")

        if not isinstance(band_list, list):
            band_list = [band_list]

        band_list = [self.band(**_) for _ in band_list]

        return {"source": source, "band_list": band_list, "lane": lane, "gel": gel}

    def band(
        self,
        elution_buffer,
        elution_volume,
        destination,
        min_bp=None,
        max_bp=None,
        band_size_range=None,
    ):
        """Helper for building band params for gel_purify

        Parameters
        ----------
        elution_buffer : str
            The type of elution buffer to be used
        elution_volume : str or Unit
            The volume of sample to be eluted
        destination : Well
            The Well the extracted samples should be eluted into
        min_bp : int, optional
            The minimum size sample to be removed.
        max_bp : int, optional
            The maximum size sample to be removed.
        band_size_range : dict, optional
            A dict of band size parameters. Should only be specified if none of
            the other band size parameters have been specified.

        Returns
        -------
        dict
            gel_purify band parameters

        Raises
        ------
        TypeError
            If destination is not a Well
        TypeError
            If `min_bp` is not an int
        TypeError
            If `max_bp` is not an int
        ValueError
            If `min_bp` is not less than `max_bp`

        """
        elution_buffer = str(elution_buffer)
        elution_volume = parse_unit(elution_volume, "microliter")
        if not isinstance(destination, Well):
            raise TypeError(f"destination: {destination} is not a Well")

        band_size_range = self._merge_param_dicts(
            dict(min_bp=min_bp, max_bp=max_bp), band_size_range
        )
        min_bp = band_size_range.get("min_bp")
        max_bp = band_size_range.get("max_bp")

        if not isinstance(min_bp, int):
            raise TypeError(f"min_bp {min_bp} was not an int")

        if not isinstance(max_bp, int):
            raise TypeError(f"max_bp {max_bp} was not an int")

        if not min_bp < max_bp:
            raise ValueError(f"min_bp: {min_bp} is not less than max_bp: {max_bp}")

        return {
            "destination": destination,
            "elution_volume": elution_volume,
            "elution_buffer": elution_buffer,
            "band_size_range": {"min_bp": min_bp, "max_bp": max_bp},
        }


# pylint: disable=redefined-builtin
class MagneticTransferBuilders(InstructionBuilders):
    """Helpers for building MagneticTransfer instruction parameters
    """

    @staticmethod
    def mag_dry(object, duration):
        """Helper for building mag_dry sub operations for MagneticTransfer

        Parameters
        ----------
        object : Container
            The Container to be operated on
        duration : str or Unit
            The duration of the operation

        Returns
        -------
        dict
            mag_dry parameters

        Raises
        ------
        TypeError
            If `object` is not a Container

        See Also
        --------
        Protocol.mag_dry

        """
        if not isinstance(object, Container):
            raise TypeError(f"object: {object} is not a Container")
        duration = parse_unit(duration, "seconds")
        return {"object": object, "duration": duration}

    @staticmethod
    def mag_incubate(object, duration, magnetize, tip_position, temperature=None):
        """Helper for building mag_incubate sub operations for MagneticTransfer

        Parameters
        ----------
        object : Container
            The Container to be operated on
        duration : str or Unit
            The duration of the operation
        magnetize : bool
            Whether the magnetic head should be engaged during the operation
        tip_position : float
            Position relative to well height that magnetic head is held
        temperature : str or Unit, optional
            The temperature of the operation

        Returns
        -------
        dict
            mag_incubate parameters

        Raises
        ------
        TypeError
            If `object` is not a Container
        TypeError
            If `magnetize` is not a bool
        ValueError
            If `tip_position` is not a positive number

        See Also
        --------
        Protocol.mag_incubate

        """
        if not isinstance(object, Container):
            raise TypeError(f"object: {object} is not a Container")
        duration = parse_unit(duration, "seconds")
        if not isinstance(magnetize, bool):
            raise TypeError(f"magnetize: {magnetize} is not a bool")
        tip_position = float(tip_position)
        if tip_position < 0:
            raise ValueError(f"tip_position: {tip_position} must be >= 0")
        if temperature is not None:
            parse_unit(temperature, "celsius")
        return {
            "object": object,
            "duration": duration,
            "magnetize": magnetize,
            "tip_position": tip_position,
            "temperature": temperature,
        }

    @staticmethod
    def mag_collect(
        object, cycles, pause_duration, bottom_position=None, temperature=None
    ):
        """Helper for building mag_collect sub operations for MagneticTransfer

        Parameters
        ----------
        object : Container
            The Container to be operated on
        cycles : int
            The number of times the operation should be repeated
        pause_duration : str or Unit
            The delay time between each repetition of the operation
        bottom_position : float, optional
            Position relative to well height where the magnetic head pauses
        temperature : str or Unit, optional
            The temperature of the operation

        Returns
        -------
        dict
            mag_collect parameters

        Raises
        ------
        TypeError
            If `object` is not a Container
        TypeError
            If `cycles` is not an int
        ValueError
            If `bottom_position` is not a positive number

        See Also
        --------
        Protocol.mag_collect

        """
        if not isinstance(object, Container):
            raise TypeError(f"object: {object} is not a Container")
        if not isinstance(cycles, int):
            raise TypeError(f"cycles: {cycles} is not an int")
        pause_duration = parse_unit(pause_duration, "seconds")
        bottom_position = float(bottom_position)
        if bottom_position < 0:
            raise ValueError(f"bottom_position: {bottom_position} must be >= 0")
        if temperature is not None:
            parse_unit(temperature, "celsius")
        return {
            "object": object,
            "cycles": cycles,
            "pause_duration": pause_duration,
            "bottom_position": bottom_position,
            "temperature": temperature,
        }

    @staticmethod
    def mag_release(
        object, duration, frequency, center=None, amplitude=None, temperature=None
    ):
        """Helper for building mag_release sub operations for MagneticTransfer

        Parameters
        ----------
        object : Container
            The Container to be operated on
        duration : str or Unit
            The duration of the operation
        frequency : str or Unit
            The frequency of the magnetic head during the operation
        center : float, optional
            Position relative to well height where oscillation is centered
        amplitude : float, optional
            Distance relative to well height to oscillate around `center`
        temperature : str or Unit, optional
            The temperature of the operation

        Returns
        -------
        dict
            mag_release parameters

        Raises
        ------
        TypeError
            If `object` is not a Container
        ValueError
            If `center` is less than 0
        ValueError
            If `amplitude` is greater than center

        See Also
        --------
        Protocol.mag_release

        """
        if not isinstance(object, Container):
            raise TypeError(f"object: {object} is not a Container")
        duration = parse_unit(duration, "seconds")
        frequency = parse_unit(frequency, "hertz")
        if center is not None:
            center = float(center)
            if center < 0:
                raise ValueError(f"center: {center} must be >= 0")
        if amplitude is not None:
            amplitude = float(amplitude)
        if center is not None and amplitude is not None and amplitude > center:
            raise ValueError(
                f"center: {center} must be greater than or equal to amplitude: "
                f"{amplitude}"
            )
        if temperature is not None:
            parse_unit(temperature, "celsius")
        return {
            "object": object,
            "duration": duration,
            "frequency": frequency,
            "center": center,
            "amplitude": amplitude,
            "temperature": temperature,
        }

    @staticmethod
    def mag_mix(
        object,
        duration,
        frequency,
        center=None,
        amplitude=None,
        magnetize=None,
        temperature=None,
    ):
        """Helper for building mag_mix sub operations for MagneticTransfer

        Parameters
        ----------
        object : Container
            The Container to be operated on
        duration : str or Unit
            The duration of the operation
        frequency : str or Unit, optional
            The frequency of the magnetic head during the operation
        center : float, optional
            Position relative to well height where oscillation is centered
        amplitude : float, optional
            Distance relative to well height to oscillate around `center`
        magnetize : bool, optional
            Whether the magnetic head should be engaged during the operation
        temperature : str or Unit, optional
            The temperature of the operation

        Returns
        -------
        dict
            mag_mix parameters

        Raises
        ------
        TypeError
            If `object` is not a Container
        ValueError
            If `center` is less than 0
        ValueError
            If `amplitude` is greater than center
        TypeError
            If `magnetize` is not a bool


        See Also
        --------
        Protocol.mag_mix

        """
        if not isinstance(object, Container):
            raise TypeError(f"object: {object} is not a Container")
        duration = parse_unit(duration, "seconds")
        frequency = parse_unit(frequency, "hertz")
        if center is not None:
            center = float(center)
            if center < 0:
                raise ValueError(f"center: {center} must be >= 0")
        if amplitude is not None:
            amplitude = float(amplitude)
        if center is not None and amplitude is not None and amplitude > center:
            raise ValueError(
                f"center: {center} must be greater than or equal to amplitude: "
                f"{amplitude}"
            )
        if magnetize is not None and not isinstance(magnetize, bool):
            raise TypeError(f"magnetize: {magnetize} is not a bool")
        if temperature is not None:
            parse_unit(temperature, "celsius")

        return {
            "object": object,
            "duration": duration,
            "frequency": frequency,
            "center": center,
            "amplitude": amplitude,
            "magnetize": magnetize,
            "temperature": temperature,
        }


class FlowCytometryBuilders(InstructionBuilders):
    """
    Builders for FlowCytometry instructions.
    """

    def __init__(self):
        super(FlowCytometryBuilders, self).__init__()
        self.excitation = None

    def laser(self, excitation, channels, power=None, area_scaling_factor=None):
        """
        Generates a dict of laser parameters.

        Parameters
        ----------
        excitation : Unit or str
            Excitation wavelength.
        channels : list(dict)
            See :meth:`FlowCytometryBuilders.channel`.
        power : Unit or str, optional
            Laser power.
        area_scaling_factor : Number, optional
            Value to scale height and area equivalently.

        Raises
        ------
        TypeError
            If `channels` is not a list of dict.
        TypeError
            If `channels` is not a list of dict.
        TypeError
            If `area_scaling_factor` is not a number.

        Returns
        -------
        dict
            A dict of laser parameters.
        """
        if not isinstance(channels, list):
            raise TypeError("channels must be a list of dict.")

        if any([not isinstance(_, dict) for _ in channels]):
            raise TypeError("channels must be a list of dict.")

        if area_scaling_factor is not None and not isinstance(
            area_scaling_factor, Number
        ):
            raise TypeError("area_scaling_factor must be a number.")

        if power is not None:
            power = parse_unit(power, "milliwatts")

        if excitation is not None:
            excitation = parse_unit(excitation, "nanometers")

        self.excitation = excitation
        channels = [self.channel(**_) for _ in channels]

        return {
            "excitation": self.excitation,
            "power": power,
            "area_scaling_factor": area_scaling_factor,
            "channels": channels,
        }

    def channel(
        self,
        emission_filter,
        detector_gain,
        measurements=None,
        trigger_threshold=None,
        trigger_logic=None,
    ):
        """
        Generates a dict of channel parameters.

        Parameters
        ----------
        emission_filter : dict
            See :meth:`FlowCytometryBuilders.emission_filter`.
        detector_gain : Unit or str
            Detector gain.
        measurements : dict, optional
            Pulse properties to record. See
            :meth:`FlowCytometryBuilders.measurements`.
        trigger_threshold : int, optional
            Channel intensity threshold. Events below this threshold.
        trigger_logic : Enum({"and", "or"}), optional
            Operator used to combine threshold.

        Raises
        ------
        TypeError
            If `trigger_threshold` is not of type int.
        ValueError
            If `trigger_logic` is not one of {"and", "or"}.

        Returns
        -------
        dict
            A dict of channel parameters.
        """
        if trigger_threshold is not None and not isinstance(trigger_threshold, int):
            raise TypeError("trigger_threshold must be of type int.")

        trigger_modes = ("and", "or")
        if trigger_logic is not None and trigger_logic not in trigger_modes:
            raise ValueError(f"trigger_logic must be one of {trigger_modes}.")

        if measurements is None:
            measurements = self.measurements()
        else:
            measurements = self.measurements(**measurements)

        emission_filter = self.emission_filter(**emission_filter)
        detector_gain = parse_unit(detector_gain, "millivolts")

        return {
            "emission_filter": emission_filter,
            "detector_gain": detector_gain,
            "measurements": measurements,
            "trigger_threshold": trigger_threshold,
            "trigger_logic": trigger_logic,
        }

    def emission_filter(self, channel_name, shortpass=None, longpass=None):
        """
        Generates a dict of emission filter parameters.

        Parameters
        ----------
        channel_name : str
            Specifies the channel name.
        shortpass : Unit or str
            Shortpass filter wavelength.
        longpass : Unit or str
            Longpass filter wavelength.

        Raises
        ------
        ValueError
            If values for longpass or shortpass are provided and
            channel_name is "FSC" or "SSC".

        Returns
        -------
        dict
            A dict of emission_filter params.
        """
        gating_modes = ("FSC", "SSC")
        if channel_name in gating_modes and (shortpass or longpass or self.excitation):
            raise ValueError(
                f"Cannot specify shortpass/longpass/excitation "
                f"parameters if channel_name is one "
                f"{gating_modes}"
            )

        if shortpass is not None:
            shortpass = parse_unit(shortpass, "nanometers")

        if longpass is not None:
            longpass = parse_unit(longpass, "nanometers")

        return {
            "channel_name": channel_name,
            "shortpass": shortpass,
            "longpass": longpass,
        }

    @staticmethod
    def measurements(area=None, height=None, width=None):
        """
        Generates a dict of measurements parameters.

        Parameters
        ----------
        area : bool, optional
            Area measurement.
        height : bool, optional
            Height measurement.
        width : bool, optional
            Width measurement.

        Raises
        ------
        TypeError
            If any of `area` | `height` | `width` are not of type bool.

        Returns
        -------
        dict
            A dict of measurements params.
        """

        if any(not isinstance(_, (bool, type(None))) for _ in (area, height, width)):
            raise TypeError("area, height, and width must be of type bool.")

        return {"area": area, "height": height, "width": width}

    def collection_conditions(
        self,
        acquisition_volume,
        flowrate,
        wait_time,
        mix_cycles,
        mix_volume,
        rinse_cycles,
        stop_criteria=None,
    ):
        """
        Generates a dict of collection_conditions parameters.

        Parameters
        ----------
        acquisition_volume : Unit or str
            Acquisition volume.
        flowrate : Unit or str
            Flow rate.
        wait_time : Unit or str
            Waiting time.
        mix_cycles : int
            Number of mixing cycles before acquisition.
        mix_volume : Unit or str
            Mixing volume.
        rinse_cycles : int
            Number of rinsing cycles.
        stop_criteria : dict, optional
            See :meth:`FlowCytometryBuilders.stop_criteria`.

        Raises
        ------
        TypeError
            If `rinse_cycles` is not of type int.
        TypeError
            If `mix_cycles` is not of type int.

        Returns
        -------
        dict
            A dict of `collection_condition` parameters.
        """
        if not isinstance(rinse_cycles, int):
            raise TypeError("rinse_cycles must be of type int.")

        if not isinstance(mix_cycles, int):
            raise TypeError("mix_cycles must be of type int.")

        acquisition_volume = parse_unit(acquisition_volume, "ul")
        wait_time = parse_unit(wait_time, "s")
        mix_volume = parse_unit(mix_volume, "ul")
        flowrate = parse_unit(flowrate, "ul/min")

        if stop_criteria is None:
            stop_criteria = self.stop_criteria(volume=acquisition_volume)
        else:
            stop_criteria = self.stop_criteria(**stop_criteria)

        return {
            "acquisition_volume": acquisition_volume,
            "flowrate": flowrate,
            "stop_criteria": stop_criteria,
            "wait_time": wait_time,
            "mix_cycles": mix_cycles,
            "mix_volume": mix_volume,
            "rinse_cycles": rinse_cycles,
        }

    @staticmethod
    def stop_criteria(volume=None, events=None, time=None):
        """
        Generates a dict of stop_criteria parameters.

        Parameters
        ----------
        volume : Unit or str, optional
            Stopping volume.
        events : int, optional
            Number of events to trigger stop.
        time : Unit or str, optional
            Stopping time.

        Raises
        ------
        TypeError
            If `events` is not of type int.

        Returns
        -------
        dict
            A dict of `stop_criteria` params.
        """
        if events is not None and not isinstance(events, int):
            raise TypeError("events must be of type int.")

        if volume is not None:
            volume = parse_unit(volume, "ul")

        if time is not None:
            time = parse_unit(time, "s")

        return {"volume": volume, "events": events, "time": time}



"""
Contains all the Autoprotocol Instruction objects

    :copyright: 2020 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

"""

# pragma pylint: disable=too-few-public-methods, redefined-builtin
from .builders import *  # pylint: disable=unused-wildcard-import
from .constants import PROVISION_MEASUREMENT_MODES


class Instruction(object):
    """Base class for an instruction that is to later be encoded as JSON.

    """

    builders = InstructionBuilders()

    def __init__(self, op, data):
        super(Instruction, self).__init__()
        self.op = op
        self.data = self._remove_empty_fields(self._remove_empty_fields(data))
        self.__dict__.update(self.data)

    def __repr__(self):
        return f"Instruction({self.op}, {self.data})"

    def _as_AST(self):
        """generates a Python object representation of Autoprotocol JSON

        Returns
        -------
        dict
            a dict of python objects that have the same structure as the
            Autoprotocol JSON for the Instruction

        Notes
        -----
        Used downstream for JSON serialization of the Instruction

        See Also
        --------
        :meth:`Protocol._refify` : Protocol serialization method

        """
        return dict(op=self.op, **self.data)

    @staticmethod
    def _remove_empty_fields(data):
        """
        Helper function to recursively search through and pop items containing
        empty dictionaries/lists or dictionaries containing fields with None
        values

        Parameters
        ----------
        data : dict or list
            Data dictionary or list to remove empty fields from

        Returns
        -------
        dict or list
            Dictionary or list without fields with None values

        """
        # We're not checking against the generic not since there are values
        # such as `0` or False which are valid.
        def filter_criteria(item):
            # Workaround for Unit equality comparison issues
            if isinstance(item, Unit):
                return False
            return item is None or item == [] or item == {}

        if isinstance(data, dict):
            return {
                k: Instruction._remove_empty_fields(v)
                for k, v in data.items()
                if not filter_criteria(v)
            }
        if isinstance(data, list):
            return [
                Instruction._remove_empty_fields(_)
                for _ in data
                if not filter_criteria(_)
            ]
        return data


class MagneticTransfer(Instruction):
    """
    A magnetic_transfer instruction is constructed as a list of lists of
    groups, executed in order, where each group is a collect, release, dry,
    incubate, or mix sub-operation.  These sub-operations control the behavior
    of tips which can be magnetized, and a heating platform. Groups in the same
    list of groups use the same tips.

    Parameters
    ----------
    groups: list(list(dict))
        dict in the groups should belong to one of the following categories:

        collect:
            Collects beads from the specified "object" by raising and lowering
            magnetized tips repeatedly with an optional pause at well bottom.
        release:
            Release beads from unmagnetized tips by oscillating the tips
            vertically into and out of the "object".
        dry:
            Dry beads on magnetized tips above and outside the "object".
        incubate:
            Incubate the "object".
        mix:
            Oscillate the tips into and out of the "object"
    magnetic_head: str
        Head-type used for this instruction
    """

    builders = MagneticTransferBuilders()

    max_objects = 8
    heads = {
        "96-deep": ["96-v-kf", "96-deep-kf", "96-deep"],
        "96-pcr": ["96-pcr", "96-v-kf", "96-flat", "96-flat-uv"],
    }

    def __init__(self, groups, magnetic_head):
        sub_ops = [subgroup for group in groups for subgroup in group]

        if not all(len(_) == 1 for _ in sub_ops):
            raise ValueError(
                f"Not all sub-operations in groups {groups} contain a single "
                f"sub-operation."
            )

        containers = {list(_.values()).pop()["object"] for _ in sub_ops}
        valid_container_types = all(
            _.container_type.shortname in self.heads[magnetic_head] for _ in containers
        )
        if not valid_container_types:
            raise ValueError(
                f"Not all containers: {containers} are in the allowed "
                f"container_types: {self.heads[magnetic_head]} for head_type: "
                f"{magnetic_head}"
            )

        # a new tip is used for each group
        if len(groups) + len(containers) > self.max_objects:
            raise RuntimeError(
                f"Only {self.max_objects} total objects can be used within the "
                f"same instruction and {len(containers)} "
                f"containers: {containers} were specified in addition to "
                f"{len(groups)} groups where each group requires a new "
                f"tip object."
            )

        magnetic_transfer = {"groups": groups, "magnetic_head": magnetic_head}

        super(MagneticTransfer, self).__init__(
            op="magnetic_transfer", data=magnetic_transfer
        )


class Dispense(Instruction):

    """
    Dispense specified reagent to specified columns. Only one of reagent,
    resource_id, and reagent_source can be specified for a given instruction.

    Parameters
    ----------
    object : Container or str
        Container for reagent to be dispensed to.
    columns : list
        Columns to be dispensed to, in the form of a list of dicts specifying
        the column number and the volume to be dispensed to that column.
        Columns are indexed from 0.
        [{"column": <column num>, "volume": <volume>}, ...]
    reagent : str, optional
        Reagent to be dispensed.
    resource_id : str, optional
        Resource to be dispensed.
    reagent_source : Well, optional
        Aliquot to be dispensed from.
    step_size : str or Unit, optional
        Specifies that the dispense operation must be executed
        using a pump that has a dispensing resolution of step_size.
    flowrate : str or Unit, optional
        The rate at which the peristaltic pump should dispense in Units of
        flow rate, e.g. microliter/second.
    nozzle_position : dict, optional
        A dict represent nozzle offsets from the center of the bottom of the
        plate's well. see Dispense.builders.nozzle_position; specified as
        {"position_x": Unit, "position_y": Unit, "position_z": Unit}.
    pre_dispense : str or Unit, optional
        The volume of reagent to be dispensed per-nozzle into waste
        immediately prior to dispensing into the ref.
    shape: dict, optional
        The shape of the dispensing head to be used for the dispense.
        See liquid_handle_builders.shape_builder; specified as
        {"rows": int, "columns": int, "format": str} with format being a
        valid SBS format.
    shake_after: dict, optional
        Parameters that specify how a plate should be shaken at the very end
        of the instruction execution.
        {"duration": Unit, "frequency": Unit, "path": str, "amplitude": Unit}
    """

    builders = DispenseBuilders()

    def __init__(
        self,
        object,
        columns,
        reagent=None,
        resource_id=None,
        reagent_source=None,
        step_size=None,
        flowrate=None,
        nozzle_position=None,
        pre_dispense=None,
        shape=None,
        shake_after=None,
    ):

        disp = {
            "object": object,
            "columns": columns,
            "reagent": reagent,
            "resource_id": resource_id,
            "reagent_source": reagent_source,
            "step_size": step_size,
            "pre_dispense": pre_dispense,
            "flowrate": flowrate,
            "nozzle_position": nozzle_position,
            "shape": shape,
            "shake_after": shake_after,
        }

        source_fields = ["reagent", "resource_id", "reagent_source"]
        sources = {_: disp[_] for _ in source_fields}
        if sum([_ is not None for _ in sources.values()]) != 1:
            raise ValueError(
                f"Exactly one of `reagent`, `resource_id`, and "
                f"`reagent_source` must be specified for Dispense, but got "
                f"{sources}."
            )

        disp = {k: v for k, v in disp.items() if v is not None}

        super(Dispense, self).__init__(op="dispense", data=disp)


class AcousticTransfer(Instruction):
    """
    Specify source and destination wells for transferring liquid via an acoustic
    liquid handler.  Droplet size is usually device-specific.

    Parameters
    ----------
    groups : list(dict)
        List of `transfer` groups in the form of:

        .. code-block:: json

            {
                "transfer": [
                    {
                        "to": "foo/A1",
                        "from": "bar/A1",
                        "volume": "1:nl"
                    }
                ]
            }

    droplet_size : str or Unit
        Volume representing a droplet_size.  The volume of each transfer should
        be a multiple of this volume.

    """

    def __init__(self, groups, droplet_size):
        super(AcousticTransfer, self).__init__(
            op="acoustic_transfer",
            data={"groups": groups, "droplet_size": droplet_size},
        )


class Spin(Instruction):
    """
    Apply the specified amount of acceleration to a plate using a centrifuge.

    Parameters
    ----------
    object : Ref or str
        Container to be centrifuged.
    acceleration : str
        Amount of acceleration to be applied to the container, expressed in
        units of "g" or "meter/second^2"
    duration : str or Unit
        Amount of time to apply acceleration.
    flow_direction : str
        Specifies the direction contents will tend toward with respect to
        the container. Valid directions are "inward" and "outward", default
        value is "inward".
    spin_direction : list(str)
        A list of "cw" (clockwise), "cww" (counterclockwise). For each
        element in the list, the container will be spun in the stated
        direction for the set "acceleration" and "duration". Default values
        are derived from the "flow_direction". If "flow_direction" is
        "outward", then "spin_direction" defaults to ["cw", "ccw"]. If
        "flow_direction" is "inward", then "spin_direction" defaults to
        ["cw"].

    """

    def __init__(
        self, object, acceleration, duration, flow_direction=None, spin_direction=None
    ):
        spin_json = {
            "object": object,
            "acceleration": acceleration,
            "duration": duration,
            "flow_direction": flow_direction,
            "spin_direction": spin_direction,
        }

        super(Spin, self).__init__(op="spin", data=spin_json)


class Thermocycle(Instruction):

    """
    Append a Thermocycle instruction to the list of instructions, with
    groups being a list of dicts in the form of:

    .. code-block:: python

        "groups": [{
            "cycles": integer,
            "steps": [{
              "duration": duration,
              "temperature": temperature,
              "read": boolean // optional (default true)
            },{
              "duration": duration,
              "gradient": {
                "top": temperature,
                "bottom": temperature
              },
              "read": boolean // optional (default true)
            }]
        }],

    To specify a melting curve, all four melting-relevant parameters must have
    a value.

    Parameters
    ----------
    object : str or Ref
        Container to be thermocycled
    groups : list(dict)
        List of thermocycling instructions formatted as above
    volume : str or Unit, optional
        Volume contained in wells being thermocycled
    dataref : str, optional
        Name of dataref representing read data if performing qPCR
    dyes : dict, optional
        Dictionary mapping dye types to the wells they're used in
    melting : dict
        Melting parameters
        See Also :meth:`Thermocycle.builders.melting`
    lid_temperature: str or Unit
        Specifies the lid temperature throughout the duration of the instruction

    Raises
    ------
    ValueError
        If one of dataref and dyes is specified but the other isn't
    ValueError
        If melting curve parameters are specified but dyes isn't
    """

    builders = ThermocycleBuilders()

    def __init__(
        self,
        object,
        groups,
        volume="25:microliter",
        dataref=None,
        dyes=None,
        melting=None,
        lid_temperature=None,
    ):

        qpcr_params = [dyes, dataref]
        if any(qpcr_params) and not all(qpcr_params):
            raise ValueError(
                f"either dyes {dyes} or dataref {dataref} was specified, but "
                f"both are required for qPCR"
            )

        if melting and any(melting.values()) and not dyes:
            raise ValueError(f"melting: {melting} was specified, but dyes was not")

        thermocycle = {
            "object": object,
            "groups": groups,
            "volume": volume,
            "dataref": dataref,
            "dyes": dyes,
            "melting": melting,
            "lid_temperature": lid_temperature,
        }

        super(Thermocycle, self).__init__(op="thermocycle", data=thermocycle)


class Incubate(Instruction):

    """
    Store a sample in a specific environment for a given duration. Once the
    duration has elapsed, the sample will be returned to the ambient environment
    until it is next used in an instruction.

    Parameters
    ----------
    object : Ref or str
        The container to be incubated
    where : Enum({"ambient", "warm_37", "cold_4", "cold_20", "cold_80"})
        Temperature at which to incubate specified container
    duration : Unit or str
        Length of time to incubate container
    shaking : bool, optional
        Specify whether or not to shake container if available at the specified
        temperature
    target_temperature : Unit or str, optional
        Specify a target temperature for a device (eg. an incubating block)
        to reach during the specified duration.
    shaking_params: dict, optional
        Specify "path" and "frequency" of shaking parameters to be used
        with compatible devices (eg. thermoshakes)
    co2 : Number, optional
        Carbon dioxide percentage

    """

    WHERE = ["ambient", "warm_30", "warm_35", "warm_37", "cold_4", "cold_20", "cold_80"]

    def __init__(
        self,
        object,
        where,
        duration,
        shaking=False,
        co2=0,
        target_temperature=None,
        shaking_params=None,
    ):
        if where not in self.WHERE:
            raise ValueError(
                "Specified `where` not contained in: " f"{', '.join(self.WHERE)}"
            )
        if where == "ambient" and shaking and not shaking_params:
            raise ValueError(
                "Shaking is only possible for ambient incubation "
                "if 'shaking_params' are specified."
            )

        incubate_json = {
            "object": object,
            "where": where,
            "duration": duration,
            "shaking": shaking,
            "co2_percent": co2,
        }
        if target_temperature:
            incubate_json["target_temperature"] = target_temperature
        if shaking_params:
            incubate_json["shaking_params"] = shaking_params
        super(Incubate, self).__init__(op="incubate", data=incubate_json)


class Agitate(Instruction):
    """
    Agitate sample(s) in a container in a specific condition for a given
    duration. Once the duration has elapsed, sample(s) will be returned
    to specified storage condition until it is used in the next
    instruction.

    Parameters
    ----------
    object : ref or str
        The container to be agitated
    mode : Enum(["vortex", "invert", "roll", "stir_bar"])
        Specifies the mode of agitation
    speed : Unit or str
        Speed to agitate container at
    duration : Unit or str
        Length of time to agitate container
    temperature : Unit or str
        Temperature to agitate container at
    mode_params : dict, optional
        Dictionary containing mode params for agitation modes
    """

    def __init__(
        self, object, mode, speed, duration, temperature=None, mode_params=None
    ):
        agitate_json = {
            "object": object,
            "mode": mode,
            "duration": duration,
            "speed": speed,
        }
        if mode_params:
            agitate_json["mode_params"] = mode_params
        if temperature:
            agitate_json["temperature"] = temperature
        super(Agitate, self).__init__(op="agitate", data=agitate_json)


class IlluminaSeq(Instruction):
    """
    Load aliquots into specified lanes for Illumina sequencing.
    The specified aliquots should already contain the appropriate mix for
    sequencing and require a library concentration reported in
    ng/uL.

    Parameters
    ----------
    flowcell : str
      Flowcell designation: "SR" or " "PE"
    lanes : list(dict)

        .. code-block:: none

          "lanes": [{
                "object": aliquot, Well,
                "library_concentration": decimal, // ng/uL
              },
              {...}]

    sequencer : str
      Sequencer designation: "miseq", "hiseq" or "nextseq"
    mode : str
      Mode designation: "rapid", "mid" or "high"
    index : str
      Index designation: "single", "dual" or "none"
    library_size: integer
        Library size expressed as an integer of basepairs
    dataref : str
        Name of sequencing dataset that will be returned.
    cycles : Enum({"read_1", "read_2", "index_1", "index_2"})
        Parameter specific to Illuminaseq read-length or number of
        sequenced bases. Refer to the ASC for more details

    """

    def __init__(
        self, flowcell, lanes, sequencer, mode, index, library_size, dataref, cycles
    ):
        seq = {
            "flowcell": flowcell,
            "lanes": lanes,
            "sequencer": sequencer,
            "mode": mode,
            "index": index,
            "library_size": library_size,
            "dataref": dataref,
            "cycles": cycles,
        }

        super(IlluminaSeq, self).__init__(op="illumina_sequence", data=seq)


class SangerSeq(Instruction):
    """
    Send the indicated wells of the container specified for Sanger sequencing.
    The specified wells should already contain the appropriate mix for
    sequencing, including primers and DNA according to the instructions
    provided by the vendor.

    Parameters
    ----------
    object : Container or str
      Container with well(s) that contain material to be sequenced.
    wells : list(str)
      Well indices of the container that contain appropriate materials to be
      sent for sequencing.
    dataref : str
      Name of sequencing dataset that will be returned.
    type: Enum({"standard", "rca"})
        Sanger sequencing type
    primer : Container, optional
        Tube containing sufficient primer for all RCA reactions.  This field
        will be ignored if you specify the sequencing type as "standard".
        Tube containing sufficient primer for all RCA reactions

    """

    def __init__(self, object, wells, dataref, type, primer=None):
        seq = {"type": type, "object": object, "wells": wells, "dataref": dataref}
        if primer and type == "rca":
            seq["primer"] = primer
        super(SangerSeq, self).__init__(op="sanger_sequence", data=seq)


class GelSeparate(Instruction):
    """
    Separate nucleic acids on an agarose gel.

    Parameters
    ----------
    objects: list or WellGroup or Well
        List of wells or WellGroup containing wells to be
        separated on gel.
    volume : str or Unit
        Volume of liquid to be transferred from each well specified to a
        lane of the gel.
    matrix : str
        Matrix (gel) in which to gel separate samples
    ladder : str
        Ladder by which to measure separated fragment size
    duration : str or Unit
        Length of time to run current through gel.
    dataref : str
        Name of this set of gel separation results.

    """

    def __init__(self, objects, volume, matrix, ladder, duration, dataref):
        super(GelSeparate, self).__init__(
            op="gel_separate",
            data={
                "objects": objects,
                "volume": volume,
                "matrix": matrix,
                "ladder": ladder,
                "duration": duration,
                "dataref": dataref,
            },
        )


class GelPurify(Instruction):

    """
    Separate nucleic acids on an agarose gel and purify.

    Parameters
    ----------
    objects: list or WellGroup
        WellGroup of wells to be purified
    volume : str or Unit
        Volume of sample required for analysis
    dataref : str
        Name of this specific dataset of measurements
    matrix : str
        Agarose concentration and number of wells on gel used for separation
    ladder : str
        Size range of ladder to be used to compare band size to
    dataref : str
        Name of dataset containing fragment sizes returned
    extract: list(dict)

        .. code-block:: none

          "extract": [{
                "elution_volume": volume,
                "elution_buffer": string, "water" | "TE",
                "lane": int,
                "band_size_range": {
                  "min_bp": int,
                  "max_bp": int,
                },
                "destination": well
              },
              {...}]

    """

    builders = GelPurifyBuilders()

    def __init__(self, objects, volume, matrix, ladder, dataref, extract):
        super(GelPurify, self).__init__(
            op="gel_purify",
            data={
                "objects": objects,
                "volume": volume,
                "matrix": matrix,
                "ladder": ladder,
                "dataref": dataref,
                "extract": extract,
            },
        )


class Absorbance(Instruction):

    """
    Read the absorbance for the indicated wavelength for the indicated
    wells. Append an Absorbance instruction to the list of instructions for
    this Protocol object.

    Parameters
    ----------
    object : str or Ref
        Object to execute the absorbance read on
    wells : list(Well) or WellGroup
        WellGroup of wells to be measured or a list of well references in
        the form of ["A1", "B1", "C5", ...]
    wavelength : str or Unit
        wavelength of light absorbance to be read for the indicated wells
    dataref : str
        name of this specific dataset of measured absorbances
    flashes : int, optional
        number of flashes for the read
    incubate_before: dict, optional
        incubation prior to reading if desired

        shaking: dict, optional
            shake parameters if desired
                amplitude: str or Unit
                    amplitude of shaking between 1 and 6:millimeter
                orbital: bool
                    True for oribital and False for linear shaking

        duration: str, Unit, optional
            time prior to plate reading

    temperature: str or Unit, optional
        set temperature to heat plate reading chamber
    settle_time: str or Unit, optional
        time to pause before each well read

    """

    builders = PlateReaderBuilders()

    def __init__(
        self,
        object,
        wells,
        wavelength,
        dataref,
        flashes=25,
        incubate_before=None,
        temperature=None,
        settle_time=None,
    ):
        json_dict = {
            "object": object,
            "wells": wells,
            "wavelength": wavelength,
            "num_flashes": flashes,
            "dataref": dataref,
            "incubate_before": incubate_before,
            "temperature": temperature,
            "settle_time": settle_time,
        }

        super(Absorbance, self).__init__(op="absorbance", data=json_dict)


class Fluorescence(Instruction):

    """
    Read the fluorescence for the indicated wavelength for the indicated
    wells.  Append a Fluorescence instruction to the list of instructions
    for this Protocol object.

    Parameters
    ----------
    object : str or Container
        object to execute the fluorescence read on
    wells : list(Well) or WellGroup
        WellGroup of wells to be measured or a list of well references in
        the form of ["A1", "B1", "C5", ...]
    excitation : str or Unit
        wavelength of light used to excite the wells indicated
    emission : str or Unit
        wavelength of light to be measured for the indicated wells
    dataref : str
        name of this specific dataset of measured absorbances
    flashes : int, optional
        number of flashes for this read
    incubate_before: dict, optional
        incubation prior to reading if desired

        shaking: dict, optional
            shake parameters if desired
                amplitude: str or Unit
                    amplitude of shaking between 1 and 6:millimeter
                orbital: bool
                    True for oribital and False for linear shaking

        duration: str, Unit, optional
            time prior to plate reading

    temperature: str or Unit, optional
        set temperature to heat plate reading chamber
    gain: float, optional
        float between 0 and 1, multiplier of maximum signal amplification
    detection_mode: str, optional
        set the detection mode of the optics, ["top", "bottom"],
        defaults to vendor specified defaults.
    position_z: dict, optional
        distance from the optics to the surface of the plate transport,
        only valid for "top" detection_mode and vendor capabilities.
        Specified as either a set distance - "manual", OR calculated from
        a WellGroup - "calculated_from_wells". Only one position_z
        determination may be specified

        .. code-block:: none

            position_z = {
                "manual": Unit
                - OR -
                "calculated_from_wells": []
            }

        manual: str, Unit, optional
            parameter available within "position_z" to set the distance from
            the optics to the plate transport.
        calculated_from_wells: list, WellGroup, Well, optional
            parameter available within "position_z" to set the distance from
            the optics to the plate transport.  If specified, the average
            optimal (maximal signal) distance will be chosen from the list
            of wells and applied to all measurements.
    settle_time: Unit, optional
        the time before the start of the measurement, defaults
        to vendor specifications
    lag_time: Unit, optional
        time between flashes and the start of the signal integration,
        defaults to vendor specifications
    integration_time: Unit, optional
        duration of the signal recording, per Well, defaults to vendor
        specifications

    """

    builders = PlateReaderBuilders()

    def __init__(
        self,
        object,
        wells,
        excitation,
        emission,
        dataref,
        flashes=25,
        incubate_before=None,
        temperature=None,
        gain=None,
        detection_mode=None,
        position_z=None,
        settle_time=None,
        lag_time=None,
        integration_time=None,
    ):
        json_dict = {
            "object": object,
            "wells": wells,
            "excitation": excitation,
            "emission": emission,
            "num_flashes": flashes,
            "dataref": dataref,
            "incubate_before": incubate_before,
            "temperature": temperature,
            "gain": gain,
            "settle_time": settle_time,
            "lag_time": lag_time,
            "integration_time": integration_time,
            "detection_mode": detection_mode,
            "position_z": position_z,
        }

        super(Fluorescence, self).__init__(op="fluorescence", data=json_dict)


class Luminescence(Instruction):

    """
    Read luminesence of indicated wells

    Parameters
    ----------
    object: str or Container
        object to execute the luminescence read on
    wells : list or WellGroup
        WellGroup or list of wells to be measured
    dataref : str
        name which dataset will be saved under
    incubate_before: dict, optional
        incubation prior to reading if desired

        shaking: dict, optional
            shake parameters if desired
                amplitude: str or Unit
                    amplitude of shaking between 1 and 6:millimeter
                orbital: bool
                    True for oribital and False for linear shaking

        duration: str, Unit, optional
            time prior to plate reading

    temperature: str or Unit, optional
        set temperature to heat plate reading chamber
    settle_time: str or Unit, optional
        time to pause before each well read
    integration_time: Unit, optional
        duration of the signal recording, per Well, defaults to vendor
        specifications

    """

    builders = PlateReaderBuilders()

    def __init__(
        self,
        object,
        wells,
        dataref,
        incubate_before=None,
        temperature=None,
        settle_time=None,
        integration_time=None,
    ):
        json_dict = {
            "object": object,
            "wells": wells,
            "dataref": dataref,
            "incubate_before": incubate_before,
            "temperature": temperature,
            "settle_time": settle_time,
            "integration_time": integration_time,
        }

        super(Luminescence, self).__init__(op="luminescence", data=json_dict)


class Seal(Instruction):

    """
    Seal indicated container using the automated plate sealer.

    Parameters
    ----------
    object : Ref or str
        Container to be sealed
    type : str, optional
        Seal type to be used (optional)
    mode : str, optional
        Method used to seal plate (optional). "thermal" or "adhesive"
    mode_params : dict, optional
        Thermal sealing parameters

        temperature : str, optional
            Temperature to seal plate at

        duration : str, optional
            Duration for which to apply heated sealing plate onto ref

    """

    def __init__(self, object, type="ultra-clear", mode=None, mode_params=None):
        seal_dict = {
            "object": object,
            "type": type,
            "mode": mode,
            "mode_params": mode_params,
        }

        super(Seal, self).__init__(op="seal", data=seal_dict)


class Unseal(Instruction):
    """
    Remove seal from indicated container using the automated plate unsealer.

    Parameters
    ----------
    object : Ref or str
        Container to be unsealed

    """

    def __init__(self, object):
        super(Unseal, self).__init__(op="unseal", data={"object": object})


class Cover(Instruction):

    """
    Place specified lid type on specified container

    Parameters
    ----------
    object : str
        Container to be covered
    lid : Enum({'standard', 'universal', 'low_evaporation'}), optional
        Type of lid to cover container with
    retrieve_lid : bool
        Flag to retrieve lid from stored location

    """

    LIDS = ["standard", "universal", "low_evaporation"]

    def __init__(self, object, lid="standard", retrieve_lid=None):
        if lid and lid not in self.LIDS:
            raise ValueError(f"{lid} is not a valid lid type")
        cover = {"object": object, "lid": lid, "retrieve_lid": retrieve_lid}

        super(Cover, self).__init__(op="cover", data=cover)


class Uncover(Instruction):
    """
    Remove lid from specified container

    Parameters
    ----------
    object : str
        Container to remove lid from
    store_lid : bool
        Flag to store the uncovered lid

    """

    def __init__(self, object, store_lid=None):
        super(Uncover, self).__init__(
            op="uncover", data={"object": object, "store_lid": store_lid}
        )


class FlowCytometry(Instruction):
    """
    This instruction provides a non-ambiguous set of parameters for the
    performance of flow cytometry.

    Parameters
    ----------
    dataref : str
        Name of dataset that will be returned.
    samples : list(Well) or Well or WellGroup
        Wells to be analyzed
    lasers : list(dict)
        See FlowCytometryBuilders.laser.
    collection_conditions : dict
        See FlowCytometryBuilders.collection_conditions.
    width_threshold : int or float, optional
        Threshold to determine width measurement.
    window_extension : int or float, optional
        Front and rear window extension.
    remove_coincident_events : bool, optional
        Remove coincident events. Defaults to false.
    """

    builders = FlowCytometryBuilders()

    def __init__(
        self,
        dataref,
        samples,
        lasers,
        collection_conditions,
        width_threshold=None,
        window_extension=None,
        remove_coincident_events=None,
    ):

        instruction = {
            "dataref": dataref,
            "samples": samples,
            "lasers": lasers,
            "collection_conditions": collection_conditions,
            "width_threshold": width_threshold,
            "window_extension": window_extension,
            "remove_coincident_events": remove_coincident_events,
        }

        super(FlowCytometry, self).__init__(op="flow_cytometry", data=instruction)


class FlowAnalyze(Instruction):

    """
    Perform flow cytometry.The instruction will be executed within the voltage
    range specified for each channel, optimized for the best sample
    separation/distribution that can be achieved within these limits. The
    vendor will specify the device that this instruction is executed on and
    which excitation and emission spectra are available. At least one negative
    control is required, which will be used to define data acquisition
    parameters as well as to determine any autofluorescent properties for the
    sample set. Additional negative positive control samples are optional.
    Positive control samples will be used to optimize single color signals and,
    if desired, to minimize bleed into other channels.


    For each sample this instruction asks you to specify the `volume` and/or
    `captured_events`. Vendors might also require `captured_events` in case
    their device does not support volumetric sample intake. If both conditions
    are supported, the vendor will specify if data will be collected only until
    the first one is met or until both conditions are fulfilled.

    Example Usage:


    Autoprotocol Output:

    Parameters
    ----------
    dataref : str
        Name of flow analysis dataset generated.
    FSC : dict
        Dictionary containing FSC channel parameters in the form of:

        .. code-block:: none

            {
              "voltage_range": {
                "low": "230:volt",
                "high": "280:volt"
                },
              "area": true,             //default: true
              "height": true,           //default: true
              "weight": false           //default: false
            }

    SSC : dict
        Dictionary of SSC channel parameters in the form of:

        .. code-block:: none

            {
              "voltage_range": {
                "low": <voltage>,
                "high": <voltage>"
                },
              "area": true,             //default: true
              "height": true,           //default: false
              "weight": false           //default: false
            }

    negative_controls : list(dict)
        List of negative control wells in the form of:

        .. code-block:: none

            {
                "well": well,
                "volume": volume,
                "captured_events": integer,     // optional, default infinity
                "channel": [channel_name]
            }

        at least one negative control is required.
    samples : list(dict)
        List of samples in the form of:

        .. code-block:: none

            {
                "well": well,
                "volume": volume,
                "captured_events": integer,     // optional, default infinity
            }

        at least one sample is required
    colors : list(dict), optional
        Optional list of colors in the form of:

        .. code-block:: none

            [{
              "name": "FitC",
              "emission_wavelength": "495:nanometer",
              "excitation_wavelength": "519:nanometer",
              "voltage_range": {
                "low": <voltage>,
                "high": <voltage>
              },
              "area": true,             //default: true
              "height": false,          //default: false
              "weight": false           //default: false
            }]


    positive_controls : list(dict), optional
        Optional list of positive control wells in the form of:

        .. code-block:: none

            [{
                "well": well,
                "volume": volume,
                "captured_events": integer,     // optional, default infinity
                "channel": [channel_name],
                "minimize_bleed": [{            // optional
                  "from": color,
                  "to": [color]
                }]
            }]

    """

    def __init__(
        self,
        dataref,
        FSC,
        SSC,
        negative_controls,
        samples,
        colors=None,
        positive_controls=None,
    ):
        flow_instr = {"dataref": dataref, "channels": {}}
        flow_instr["channels"]["FSC"] = FSC
        flow_instr["channels"]["SSC"] = SSC
        flow_instr["negative_controls"] = negative_controls
        flow_instr["samples"] = samples
        if colors:
            flow_instr["channels"]["colors"] = colors
        if positive_controls:
            flow_instr["positive_controls"] = positive_controls

        super(FlowAnalyze, self).__init__(op="flow_analyze", data=flow_instr)


class Oligosynthesize(Instruction):
    """
    Parameters
    ----------
    oligos : list of dicts
        List of oligonucleotides to synthesize.  Each dictionary should
        contain the oligo's sequence, destination, scale and purification

        .. code-block:: none

            [
                {
                  "destination": "my_plate/A1",
                  "sequence": "GATCRYMKSWHBVDN",
                    // - standard IUPAC base codes
                    // - IDT also allows rX (RNA), mX (2' O-methyl RNA), and
                    //   X*/rX*/mX* (phosphorothioated)
                    // - they also allow inline annotations for modifications,
                    //   eg "GCGACTC/3Phos/" for a 3' phosphorylation
                    //   eg "aggg/iAzideN/cgcgc" for an internal modification
                  "scale": "25nm" | "100nm" | "250nm" | "1um",
                  "purification": "standard" | "page" | "hplc",
                    // default: standard
                },
                ...
            ]
    """

    def __init__(self, oligos):
        super(Oligosynthesize, self).__init__(
            op="oligosynthesize", data={"oligos": oligos}
        )


class Autopick(Instruction):
    """
    Pick colonies from the agar-containing location(s) specified in `sources`
    to the location(s) specified in `dests` in highest to lowest rank order
    until there are no more colonies available.  If fewer than min_abort
    pickable colonies have been identified from the location(s) specified in
    `sources`, the run will stop and no further instructions will be executed.

    Parameters
    ----------
    groups : list(dict)
        Groups of colonies to pick and where to transport them to
    criteria : dict
        Dictionary of autopicking criteria.
    dataref: str
        Name of dataset to save the picked colonies to

    """

    def __init__(self, groups, criteria, dataref):
        pick = {"groups": groups, "dataref": dataref, "criteria": criteria}

        super(Autopick, self).__init__(op="autopick", data=pick)


class ImagePlate(Instruction):
    """
    Capture an image of the specified container.

    Parameters
    ----------
    object : str
        Container to take image of
    mode : str
        Imaging mode (currently supported: "top")
    dataref : str
        Name of data reference of resulting image

    """

    def __init__(self, object, mode, dataref):
        super(ImagePlate, self).__init__(
            op="image_plate", data={"object": object, "mode": mode, "dataref": dataref}
        )


class Image(Instruction):
    """
    Capture an image of the specified container.

    Parameters
    ----------
    ref : Container
        Container of which to take image.
    mode : str
        Angle of image, one of "top", "bottom", "side"
    num_images : int
        Number of images taken of the container. Defaults to 1.
    dataref : str
        Name of data reference of resulting image
    backlighting : Bool, optional
        Whether back-lighting is desired.
    magnification : float
        Ratio of sizes of the image projected on the camera
        sensor compared to the actual size of the object
        captured. Defaults to 1.0.
    exposure : dict, optional
        Parameters to control exposure: "aperture", "iso",
        and "shutter_speed".

        shutter_speed: Unit, optional
            Duration that the imaging sensor is exposed.
        iso : Float, optional
            Light sensitivity of the imaging sensor.
        aperture: Float, optional
            Diameter of the lens opening.

    """

    def __init__(
        self, ref, mode, dataref, num_images, backlighting, exposure, magnification
    ):
        json_dict = {
            "object": ref,
            "mode": mode,
            "dataref": dataref,
            "num_images": num_images,
            "magnification": magnification,
            "back_lighting": backlighting,
            "exposure": exposure,
        }

        super(Image, self).__init__(op="image", data=json_dict)


class Provision(Instruction):
    """
    Provision a commercial resource from a catalog into the specified
    destination well(s).  A new tip is used for each destination well
    specified to avoid contamination.

    Parameters
    ----------
    resource_id : str
      Resource ID from catalog.
    dests : list(dict)
      Destination(s) for specified resource, together with volume information
    measurement_mode : str
      Measurement mode. Possible values are :py:class:`autoprotocol.constants.MEASUREMENT_MODES`

    Raises
    ------
    TypeError
      If resource_id is not a string.
    RuntimeError
      If length of the list of volumes specified does not match the number of
      destination wells specified.
    RuntimeError
      If the measurement mode is not supported.
    TypeError
      If volume is not specified as a string or Unit (or a list of either)

    """

    def __init__(self, resource_id, dests, measurement_mode="volume"):
        if measurement_mode not in PROVISION_MEASUREMENT_MODES:
            raise RuntimeError(
                f"{measurement_mode} is not a valid measurement mode for provisioning"
            )

        super(Provision, self).__init__(
            op="provision",
            data={
                "resource_id": resource_id,
                "measurement_mode": measurement_mode,
                "to": dests,
            },
        )


class FlashFreeze(Instruction):
    """
    Flash freeze the contents of the specified container by submerging it in
    liquid nitrogen for the specified amount of time.

    Parameters
    ----------
    object : Container or str
        Container to be flash frozen.
    duration : str or Unit
        Duration to submerge specified container in liquid nitrogen.

    """

    def __init__(self, object, duration):
        super(FlashFreeze, self).__init__(
            op="flash_freeze", data={"object": object, "duration": duration}
        )


class Evaporate(Instruction):
    """
    Removes liquid or moisture from sample.

    Parameters
    ----------
    ref : Container
        Sample container
    mode : str
        Mode of evaporation
    duration : Unit or str
        Duration object is processed
    evaporator_temperature : Unit or str
        Temperature object is exposed to
    mode_params : dict
        Dictionary of parameters for evaporation mode
    """

    builders = EvaporateBuilders()

    def __init__(self, ref, mode, duration, evaporator_temperature, mode_params):
        json_dict = {
            "object": ref,
            "mode": mode,
            "duration": duration,
            "evaporator_temperature": evaporator_temperature,
            "mode_params": mode_params,
        }
        super(Evaporate, self).__init__(op="evaporate", data=json_dict)


class SPE(Instruction):

    """
    Apply a solid phase extraction (spe) technique to a sample.

    Parameters
    ----------
    well : Well
        Well to solid phase extract.
    cartridge : str
        Cartridge to use for solid phase extraction.
    pressure_mode : str
        The direction of pressure applied to the cartridge to force
        liquid flow. One of "positive", "negative".
    load_sample: dict
        Parameters for applying the sample to the cartridge.
        Single 'mobile_phase_param'.
    elute: list(dict)
        Parameters for applying a mobile phase to the cartridge
        with one or more solvents. List of 'mobile_phase_params'.
        Requires `destination_well`.
    condition: list(dict), optional
        Parameters for applying a mobile phase to the cartridge
        with one or more solvents. List of 'mobile_phase_params'.
    equilibrate: list(dict), optional
        Parameters for applying a mobile phase to the cartridge
        with one or more solvents. List of 'mobile_phase_params'.
    rinse: list(dict), optional
        Parameters for applying a mobile phase to the cartridge
        with one or more solvents. List of 'mobile_phase_params'.

        mobile_phase_params:
            resource_id: str
                Resource ID of desired solvent.
            volume: volume
                Volume added to the cartridge.
            loading_flowrate: Unit
                Speed at which volume is added to cartridge.
            settle_time: Unit
                Duration for which the solvent remains on the cartridge
                before a pressure mode is applied.
            processing_time: Unit
                Duration for which pressure is applied to the cartridge
                after `settle_time` has elapsed.
            flow_pressure: Unit
                Pressure applied to the column.
            destination_well: Well
                Destination well for eluate.  Required parameter for
                each `elute` mobile phase parameter

    """

    builders = SPEBuilders()

    def __init__(
        self,
        well,
        cartridge,
        pressure_mode,
        load_sample,
        elute,
        condition,
        equilibrate,
        rinse,
    ):
        json_dict = {
            "object": well,
            "cartridge": cartridge,
            "pressure_mode": pressure_mode,
            "condition": condition,
            "load_sample": load_sample,
            "rinse": rinse,
            "elute": elute,
            "equilibrate": equilibrate,
        }

        super(SPE, self).__init__(op="spe", data=json_dict)


class MeasureConcentration(Instruction):
    """
    Measure the concentration of DNA, ssDNA, RNA or Protein in the specified
    volume of the source aliquots.

    Parameters
    ----------
    object : list or WellGroup
        WellGroup of wells to be measured
    volume : str or Unit
        Volume of sample required for analysis
    dataref : str
        Name of this specific dataset of measurements
    measurement : str
        Class of material to be measured. One of ["DNA", "ssDNA", "RNA",
        "protein"].

    """

    def __init__(self, object, volume, dataref, measurement):
        json_dict = {
            "object": object,
            "volume": volume,
            "dataref": dataref,
            "measurement": measurement,
        }
        super(MeasureConcentration, self).__init__(
            op="measure_concentration", data=json_dict
        )


class Sonicate(Instruction):
    """
    Sonicate wells using high intensity ultrasonic vibrations.

    Parameters
    ----------
    wells : Well or WellGroup or list(Well)
       Wells to be sonicated
    duration : Unit or str
        Duration for which to sonicate wells
    mode: Enum({"bath", "horn"})
        Sonicating method to be used, must be "horn" or "bath". Sonicate
        mode "horn" uses metal probe to create a localized shear force
        directly in the sample media; "bath" mode applies ultrasound to
        wells held inside a bath.
    temperature: Unit or str, optional
        Temperature at which the sample is kept during sonication. Optional,
        defaults to ambient
    frequency: Unit or str, optional
        Frequency of the ultrasonic wave, usually indicated in kHz.
        Optional; defaults to the most commonly used frequency for each
        mode: 20 kHz for `horn`, and 40 kHz for `bath` mode
    mode_params: Dict
        Dictionary containing mode parameters for the specified mode.

    """

    def __init__(self, wells, duration, mode, mode_params, frequency, temperature):
        json_dict = {
            "wells": wells,
            "duration": duration,
            "frequency": frequency,
            "mode": mode,
            "mode_params": mode_params,
        }
        if temperature:
            json_dict["temperature"] = temperature
        super(Sonicate, self).__init__(op="sonicate", data=json_dict)


class MeasureMass(Instruction):
    """
    Measure the mass of containers

    Parameters
    ----------
    object : Container
        Container ref
    dataref: str
        Name of the data for the measurement
    """

    def __init__(self, object, dataref):
        json_dict = {"object": object, "dataref": dataref}
        super(MeasureMass, self).__init__(op="measure_mass", data=json_dict)


class MeasureVolume(Instruction):
    """
    Measure the mass of containers

    Parameters
    ----------
    object: list(Container)
        list of containers
    dataref: str
        Name of the data for the measurement
    """

    def __init__(self, object, dataref):
        json_dict = {"object": object, "dataref": dataref}
        super(MeasureVolume, self).__init__(op="measure_volume", data=json_dict)


class CountCells(Instruction):
    """
    Count the number of cells in a sample that are positive/negative
    for a given set of labels.

    Parameters
    ----------
    wells: WellGroup
        List of wells that will be used for cell counting.
    volume: Unit
        Volume that should be consumed from each well for the purpose
        of cell counting.
    dataref: str
        Name of dataset that will be returned.
    labels: list(string), optional
        Cells will be scored for presence or absence of each label
        in this list. If staining is required to visualize these labels,
        they must be added before execution of this instruction.

    """

    def __init__(self, wells, volume, dataref, labels=None):
        json_dict = {
            "wells": wells,
            "volume": volume,
            "dataref": dataref,
            "labels": labels,
        }

        super(CountCells, self).__init__(op="count_cells", data=json_dict)


class Spectrophotometry(Instruction):
    """
    Execute a Spectrophotometry plate read on the obj.

    Parameters
    ----------
    dataref : str
        Name of the resultant dataset to be returned.
    object : Container or str
        Container to be read.
    groups : list
        A list of groups generated by SpectrophotometryBuilders groups builders,
        any of absorbance_mode_params, fluorescence_mode_params,
        luminescence_mode_params, or shake_mode_params.
    interval : Unit or str, optional
        The time between each of the read intervals.
    num_intervals : int, optional
        The number of times that the groups should be executed.
    temperature : Unit or str, optional
        The temperature that the entire instruction should be executed at.
    shake_before : dict, optional
        A dict of params generated by SpectrophotometryBuilders.shake_before
        that dictates how the obj should be incubated with shaking before any of
        the groups are executed.
    """

    builders = SpectrophotometryBuilders()

    def __init__(
        self,
        dataref,
        object,
        groups,
        interval=None,
        num_intervals=None,
        temperature=None,
        shake_before=None,
    ):
        spec = {
            "dataref": dataref,
            "object": object,
            "groups": groups,
            "interval": interval,
            "num_intervals": num_intervals,
            "temperature": temperature,
            "shake_before": shake_before,
        }

        super(Spectrophotometry, self).__init__(op="spectrophotometry", data=spec)


class LiquidHandle(Instruction):
    """Manipulates liquids within locations

    A liquid handle instruction is constructed as a list of locations, where
    each location consists of the well location and the tip transports carried
    out within the well.

    Each liquid handle instruction corresponds to a single tip or set of tips.

    Parameters
    ----------
    locations : list(dict)
        See Also :meth:`LiquidHandle.builders.location`
    shape : dict, optional
        See Also :meth:`LiquidHandle.builders.shape`
    mode : str, optional
        the liquid handling mode
    mode_params : dict, optional
        See Also :meth:`LiquidHandle.builders.instruction_mode_params`
    """

    builders = LiquidHandleBuilders()

    def __init__(self, locations, shape=None, mode=None, mode_params=None):
        data = {
            "locations": locations,
            "shape": shape,
            "mode": mode,
            "mode_params": mode_params,
        }

        super(LiquidHandle, self).__init__(op="liquid_handle", data=data)



#!/usr/bin/env python3
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

# Load version
exec(open("autoprotocol/version.py").read())  # pylint: disable=exec-used

# Test Runner (reference: https://docs.pytest.org/en/latest/goodpractices.html)
class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = "--cov=autoprotocol --cov-report=term"  # pylint: disable=attribute-defined-outside-init

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


# Test and Documentation dependencies
test_deps = [
    "coverage>=4.5, <5",
    "pre-commit>=2.4, <3",
    "pylint==2.5.2",  # should be consistent with .pre-commit-config.yaml
    "pytest>=5.4, <6",
    "pytest-cov>=2, !=2.8.1",
    "tox>=3.15, <4",
]

doc_deps = [
    "releases>=1.6.3, <2",
    "Sphinx>=2.4, <3",
    "sphinx_rtd_theme>=0.4.3, <1",
    "semantic-version==2.6.0",
]


setup(
    name="autoprotocol",
    url="https://github.com/autoprotocol/autoprotocol-python",
    maintainer="The Autoprotocol Development Team",
    description="Python library for generating Autoprotocol",
    license="BSD",
    maintainer_email="support@strateos.com",
    version=__version__,  # pylint: disable=undefined-variable
    install_requires=["Pint==0.9"],
    python_requires=">=3.6",
    tests_require=test_deps,
    extras_require={"docs": doc_deps, "test": test_deps},
    cmdclass={"pytest": PyTest},
    packages=["autoprotocol", "autoprotocol.liquid_handle"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
